/*
 * PROPRIETARY INFORMATION.  This software is proprietary to POWDER
 * Development, and is not to be reproduced, transmitted, or disclosed
 * in any way without written permission.
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        ai.cpp ( Save Scummer Library, C++ )
 *
 * COMMENTS:
 */

#include "mob.h"

#include "map.h"
#include "msg.h"
#include "text.h"
#include "speed.h"
#include "item.h"
#include "avatar.h"

///
/// MINDSTATE Implementation
///

MINDSTATE::MINDSTATE()
{
    ACTIVITY_NAMES	activity;
    
    myAttention = 0;
    FOREACH_ACTIVITY(activity)
    {
	myActivity[activity] = 0;
    }

    // A nice safe distance
    myWaterDist = 10;

    // Number of times I've been yelled at.
    myScolds = 0;
}

void
MINDSTATE::load(istream &is)
{
    is.read((char *) &myAttention, sizeof(int));
    is.read((char *) &myScolds, sizeof(int));
    is.read((char *) &myWaterDist, sizeof(int));
    is.read((char *) &myTagger, sizeof(MOB_NAMES));

    ACTIVITY_NAMES	activity;
    FOREACH_ACTIVITY(activity)
    {
	is.read((char *) &myActivity[activity], sizeof(int));
    }
}

void
MINDSTATE::save(ostream &os) const
{
    os.write((const char *) &myAttention, sizeof(int));
    os.write((const char *) &myScolds, sizeof(int));
    os.write((const char *) &myWaterDist, sizeof(int));
    os.write((const char *) &myTagger, sizeof(MOB_NAMES));

    ACTIVITY_NAMES	activity;
    FOREACH_ACTIVITY(activity)
    {
	os.write((const char *) &myActivity[activity], sizeof(int));
    }
}
    
void
MINDSTATE::payAttention(ACTIVITY_NAMES state, int attention)
{
    myAttention += attention;

    myActivity[state] += attention;
}

ACTIVITY_NAMES
MINDSTATE::favoriteActivity() const
{
    ACTIVITY_NAMES	result = ACTIVITY_TAG;
    ACTIVITY_NAMES	activity;

    FOREACH_ACTIVITY(activity)
    {
	if (myActivity[activity] > myActivity[result])
	    result = activity;
    }
    return result;
}

bool
MOB::aiForcedAction()
{
    // Check for the phase...
    PHASE_NAMES		phase;

    phase = spd_getphase();

    // On normal & slow phases, items get heartbeat.
    if (phase == PHASE_NORMAL || phase == PHASE_SLOW)
    {
	ITEM	*item, *next;
	for (item = myInventory; item; item = next)
	{
	    next = item->getNext();

	    if (item->runHeartbeat())
	    {
		removeItem(item);
		delete item;
	    }
	}

	// Regenerate if we are capable of it
	if (isRegenerating())
	{
	    gainHP(1);
	}

	// Apply damage if inside.
	if (isSwallowed())
	{
	    msg_format("%S <be> digested!", this);
	    if (applyDamage(0, 1))
		return true;
	}
    }

    switch (phase)
    {
	case PHASE_FAST:
	    // Not fast, no phase
	    if (isFast())
		return false;
	    return true;

	case PHASE_QUICK:
	    // Not quick, no phase!
	    if (isQuick())
		return false;
	    return true;

	case PHASE_SLOW:
	    if (isSlow())
		return true;
	    break;

	case PHASE_NORMAL:
	    break;
    }
    return false;
}

bool
MOB::aiDoAI()
{
    // Rebuild our home location.
    if (myHX < 0)
    {
	myHX = getX();
	myHY = getY();
    }

    if (myFleeCount > 0)
    {
	myFleeCount--;
	if (!isAvatar() && aiFleeFromAvatar())
	    return true;
    }

    myBoredom++;
    if (myBoredom > 1000)
	myBoredom = 1000;

    switch (getAI())
    {
	case AI_NONE:
	    // Just stay put!
	    return true;

	case AI_STAYHOME:
	    // If we are at home stay put.
	    if (getX() == myHX && getY() == myHY)
		return true;

	    // Otherwise, go home.
	    // FALL THROUGH

	case AI_HOME:
	    // Move towards our home square.
	    int		dist;

	    dist = MAX(ABS(getX()-myHX), ABS(getY() - myHY));

	    // If we can shoot the avatar or charge him, do so.
	    if (aiRangeAttack())
		return true;
	    if (aiCharge(MOB::getAvatar(), AI_CHARGE))
		return true;

	    // A good chance to just stay put.
	    if (rand_chance(70))
	    {
		return actionBump(0, 0);
	    }

	    if (rand_choice(dist))
	    {
		// Try to home.
		if (aiMoveTo(myHX, myHY))
		    return true;
	    }
	    // Default to wandering
	    return aiRandomWalk();

	case AI_ORTHO:
	    if (aiRangeAttack())
		return true;
	    if (aiCharge(MOB::getAvatar(), AI_CHARGE, true))
		return true;
	    // Default to wandering
	    return aiRandomWalk(true);

	case AI_CHARGE:
	    if (aiRangeAttack())
		return true;
	    if (aiCharge(MOB::getAvatar(), AI_CHARGE))
		return true;
	    // Default to wandering
	    return aiRandomWalk();

	case AI_PATHTO:
	    if (aiKillPathTo(MOB::getAvatar()))
		return true;
		    
	    // Default to wandering
	    return aiRandomWalk();

	case AI_FLANK:
	    if (aiRangeAttack())
		return true;
	    if (aiCharge(MOB::getAvatar(), AI_FLANK))
		return true;
	    // Default to wandering
	    return aiRandomWalk();

	case AI_RANGECOWARD:
	    // If we can see avatar, shoot at them!
	    // We keep ourselves at range if possible, unless
	    // retreat is blocked, in which case we will melee.
	    if (getMap()->isFOV(getX(), getY()))
	    {
		int		 dist, dx, dy;

		if (aiAcquireAvatar())
		{
		    dx = SIGN(myTX - getX());
		    dy = SIGN(myTY - getY());
		    dist = MAX(ABS(getX()-myTX), ABS(getY() - myTY));

		    if (dist == 1)
		    {
			// We are in melee range.  Try to flee.
			if (aiFleeFrom(myTX, myTY))
			    return true;

			// Failed to flee.  Kill!
			return actionMelee(dx, dy);
		    }

		    // Try ranged attack.
		    if (aiRangeAttack())
			return true;

		    // Failed range attack.  If outside of range, charge.
		    if (dist > getRangedRange())
		    {
			if (aiCharge(MOB::getAvatar(), AI_FLANK))
			    return true;
		    }

		    // Otherwise, wander within the range hoping to line up.
		    // Trying to double think the human is likely pointless
		    // as they'll be trying to avoid lining up to.
		    return aiRandomWalk();
		}
	    }
		
	    if (aiCharge(MOB::getAvatar(), AI_FLANK))
		return true;

	    // Default to wandering
	    return aiRandomWalk();

	    
	case AI_COWARD:
	    if (aiRangeAttack())
		return true;

	    if (aiFleeFromAvatar())
		return true;
	    // Failed to flee, cornered, so charge!
	    if (aiCharge(MOB::getAvatar(), AI_CHARGE))
		return true;

	    return aiRandomWalk();

	case AI_MOUSE:
	    return aiDoMouse();

	case AI_STRAIGHTLINE:
	    return aiStraightLine();

	case AI_CHILD:
	    return aiChild();
    }

    return false;
}

ITEM *
MOB::aiFindCloseMappedItem() const
{
    ITEM		*item, *minitem = 0;
    int			 mindist = 1000, dist;

    for (item = getMap()->getItemHead(); item; item = item->getNext())
    {
	// Only count mapped items we know about.
	if (!getMap()->isMapped(item->getX(), item->getY()))
	    continue;

	// Ignore items we can't reach.
	if (!getMap()->isConnected(item->getPos(), getPos(), myMind.myWaterDist))
	    continue;

	dist = MAP::dist(this, item->getX(), item->getY());

	if (!minitem || dist < mindist)
	{
	    minitem = item;
	    mindist = dist;
	}
    }
    return minitem;
}

MOB *
MOB::aiFindCloseChild(MOB_NAMES avoid) const
{
    MOB			*biggestfoe = 0, *foe;
    
    for (foe = getMap()->getMobHead(); foe; foe = foe->getNext())
    {
	if (!foe->defn().ischild)
	    continue;

	if (foe == this)
	    continue;

	if (getMap()->isConnected(this->getPos(), foe->getPos(), myMind.myWaterDist))
	{
	    // If biggest foe is the avoid mob, skip distance check
	    if (biggestfoe && (biggestfoe->getDefinition() != avoid))
		if (MAP::dist(this, biggestfoe) < MAP::dist(this, foe))
		    continue;
	    // If we are the avoid check and there is a biggest foe, skip.
	    if (biggestfoe && (foe->getDefinition() == avoid))
		continue;
		    
	    biggestfoe = foe;
	}
    }
    return biggestfoe;
}

ITEM *
MOB::aiFindCloseRock() const
{
    ITEM		*item, *minitem = 0;
    int			 mindist = 1000, dist;

    for (item = getMap()->getItemHead(); item; item = item->getNext())
    {
	// Only count mapped items we know about.
	if (!getMap()->isMapped(item->getX(), item->getY()))
	    continue;

	// Ignore items we can't reach.
	if (!getMap()->isConnected(item->getPos(), getPos(),
				myMind.myWaterDist))
	    continue;

	// Ignore items who are in too deep of water.
	if (getMap()->waterDepth(item->getPos()) < myMind.myWaterDist)
	    continue;

	// Ignore items that are good enough
	if (aiIsRockPosGood(item->getPos()))
	    continue;

	// Ignore items too big to pick up.
	if (item->getSize() > getSize())
	    continue;

	// Ignore rocks which the placer decided was final.
	if (item->isFinalPos())
	    continue;

	dist = MAP::dist(this, item->getX(), item->getY());

	if (!minitem || dist < mindist)
	{
	    minitem = item;
	    mindist = dist;
	}
    }
    return minitem;
}

bool
MOB::aiIsRockPosGood(POS p) const
{
    MOB		*avatar = getAvatar();

    if (!avatar)
	return false;		// Nothing good anymore :<

    // 5 squares is close enough.
    if (getMap()->getDist(avatar->getPos(), p) < 5)
    {
	return true;
    }

    return false;
}

void
MOB::aiChooseRockDrop(int &x, int &y) const
{
    MOB		*avatar = getAvatar();

    if (!avatar)
    {
	// In sadness, just drop it here.
	x = getX();
	y = getY();
	return;
    }
    // For now, bring it to the avatar.
    x = avatar->getX();
    y = avatar->getY();
}

MOB *
MOB::aiFindIt() const
{
    MOB			*foe;
    
    for (foe = getMap()->getMobHead(); foe; foe = foe->getNext())
    {
	if (foe->myAIState == AI_STATE_TAGIT)
	    return foe;
    }
    return 0;
}

MOB *
MOB::aiFindEnemy() const
{
    PTRLIST<MOB *>	list;
    int			i;
    MOB			*biggestfoe, *foe;
    
    // The problem with this is if we have two stationary foes
    // a and b, where we hate b more, if b is out of sight and
    // we step to a and we then step to b, but in doing so hide
    // b, so we return our attention again to a.  An infinite loop
    // then develops.
    getVisibleEnemies(list);
    biggestfoe = 0;
    for (i = 0; i < list.entries(); i++)
    {
	foe = list(i);
	biggestfoe = aiHateMore(biggestfoe, foe);
    }
    return biggestfoe;
}

MOB *
MOB::aiHateMore(MOB *a, MOB *b) const
{
    // Check which is closer.
    int		bdx, bdy, adx, ady;
    int		adist, bdist;
    bool	alinedup, blinedup;

    // Trivial cases.
    if (!a)
	return b;
    if (!b)
	return a;

    adx = a->getX() - getX();
    ady = a->getY() - getY();
    bdx = b->getX() - getX();
    bdy = b->getY() - getY();

    adist = MAP::dist(a, this);
    bdist = MAP::dist(b, this);

    // If a is melee, take it.
    if (adist <= 1)
    {
	if (bdist <= 1)
	{
	    // Both are close.  We need to take out the one projected
	    // to do the most damage over the length of time it is
	    // expected we could kill it.
	    double		admg, bdmg, ourdmg;
	    double		acost, bcost;

	    ourdmg = getMeleeDPDF().expectedValue();
	    admg = a->getMeleeDPDF().expectedValue();
	    bdmg = b->getMeleeDPDF().expectedValue();
	    
	    // We compare the opportunity costs of killing either
	    // a or b first.
	    // It will take a->getHP() / ourdmg turns to kill a.
	    // In that time, b will do (bdmg * a->getHP() / ourdmg)
	    // damage to us.  Since we want to minimize damage, we pick
	    // the enemy for whom this value is highest.
	    // (Thanks to Thijs van Ommen who corrected my earlier
	    // calculation)
	    // TODO: The big exception is if the enemy can regenerate.
	    // Normal regenration should be modelled by lowering our damage
	    // output, vampirism modelled by subtracting admg from ourdmg.
	    acost = (admg * b->getHP()) / ourdmg;
	    bcost = (bdmg * a->getHP()) / ourdmg;

	    if (acost > bcost)
		return a;
	    return b;
	}
	return a;
    }
    // Ditto for b.
    if (bdist <= 1)
	return b;

    alinedup = MAP::isLinedUp(a, this);
    blinedup = MAP::isLinedUp(b, this);

    // Take the one that is lined up for attack
    if (alinedup)
    {
	if (blinedup)
	{
	    // Take closer..
	    if (adist < bdist)
		return a;
	    return b;
	}
	// Pick the lined up one.
	return a;
    }
    if (blinedup)
	return b;

    // Take closer
    if (adist < bdist)
    {
	return a;
    }
    return b;
}

bool
MOB::aiDoMouse()
{
    int		lastdir, dir, newdir;
    int		dx, dy;
    bool	found = false;
    int		i;

    // Least three bits track our last direction.
    lastdir = myAIData & 7;

    // We want to try and track a wall.  We want this wall to
    // be on our right hand side.

    // Take a look at right hand square.
    rand_angletodir((lastdir + 2) & 7, dx, dy);
    
    if (!canMove(getX() + dx, getY() + dy, true) || 
	getMap()->getMob(getX() + dx, getY() + dy))
    {
	// Got a wall to the right.  Try first available direction.
	// Because we want to hug using diagonals, we try the forward
	// and right first.
	dir = (lastdir+1) & 7;
    }
    else
    {
	// No wall on right.  Try and go straight forward by default.
	dir = lastdir;
    }

    // If everyway is blocked, bump straight forward.
    newdir = lastdir;
    for (i = 0; i < 8; i++)
    {
	rand_angletodir(dir, dx, dy);
	if (canMove(getX() + dx, getY() + dy, true) &&
	    !getMap()->getMob(getX() + dx, getY() + dy))    
	{
	    newdir = dir;
	    break;
	}
	// Keeping wall on right means turning left first!
	dir--;
	dir &= 7;
    }

    if (newdir == -1)
	newdir = lastdir;
    else
	found = true;

    // Store our last direction.
    myAIData &= ~7;
    myAIData |= newdir;

    if (found)
    {
	return actionBump(dx, dy);
    }

    // Mouse is cornered!  Return to normal AI.
    return false;
}

bool
MOB::aiStraightLine()
{
    int		lastdir;
    int		dx, dy;
    MOB		*target;

    // Least three bits track our last direction.
    lastdir = myAIState & 7;

    // Bump & go.

    // Take a look at right hand square.
    rand_angletodir(lastdir, dx, dy);
    
    // If target is avatar, run into him!
    target = getMap()->getMob(getX() + dx, getY() + dy);
    if (target && !isFriends(target))
    {
	return actionBump(dx, dy);
    }
    
    if (!canMove(getX() + dx, getY() + dy, true))
    {
	// Can't go there!  Pick new direction.
	// Do not want same direction again.
	lastdir += rand_choice(7) + 1;
	lastdir &= 7;
    }
    myAIData = lastdir;
    rand_angletodir(lastdir, dx, dy);
    
    // If target is avatar, run into him!
    target = getMap()->getMob(getX() + dx, getY() + dy);
    if (target && !isFriends(target))
    {
	return actionBump(dx, dy);
    }

    // Bump failed.
    if (!canMove(getX() + dx, getY() + dy, true))
	return false;

    // Move forward!
    return actionBump(dx, dy);
}

bool
MOB::aiFleeFromAvatar()
{
    if (aiAcquireAvatar())
    {
	if (aiFleeFrom(myTX, myTY))
	    return true;
    }

    return false;
}

bool
MOB::aiFleeFrom(int x, int y)
{
    int		dx, dy;
    int		angle, resultangle = 0, i;
    int		choice = 0;

    dx = -SIGN(x - getX());
    dy = -SIGN(y - getY());
    
    angle = rand_dirtoangle(dx, dy);

    // 3 ones in same direction get equal chance.
    for (i = angle-1; i <= angle+1; i++)
    {
	rand_angletodir(i, dx, dy);
	if (canMove(getX()+dx, getY()+dy))
	{
	    choice++;
	    if (!rand_choice(choice))
		resultangle = i;
	}
    }

    if (!choice)
    {
	// Try a bit more desperately...
	for (i = angle-2; i <= angle+2; i += 4)
	{
	    rand_angletodir(i, dx, dy);
	    if (canMove(getX()+dx, getY()+dy))
	    {
		choice++;
		if (!rand_choice(choice))
		    resultangle = i;
	    }
	}
    }

    if (choice)
    {
	// Move in the direction
	rand_angletodir(resultangle, dx, dy);
	return actionBump(dx, dy);
    }

    // Failed
    return false;
}

bool
MOB::aiAcquireAvatar()
{
    MOB		*a;
    a = MOB::getAvatar();

    return aiAcquireTarget(a);
}

bool
MOB::aiAcquireTarget(MOB *a)
{
    if (a && a->isSwallowed())
    {
	myTX = -1;
	return false;
    }
    // If we can see avatar, charge at them!  Otherwise, move
    // to our target if set.  Otherwise random walk.
    if (a && 
	getMap()->isFOV(getX(), getY()) &&
	getMap()->isFOV(a->getX(), a->getY()))
    {
	int		dist;
	// Determine distance
	dist = MAX(ABS(a->getX() - getX()), ABS(a->getY() - getY()));
	if (dist <= defn().sightrange)
	{
	    myTX = a->getX();
	    myTY = a->getY();
	}
    }
    else
    {
	// Random chance of forgetting avatar location.
	// Also, if current location is avatar location, we forget it.
	if (getX() == myTX && getY() == myTY)
	    myTX = -1;
	if (!rand_choice(10))
	    myTX = -1;
    }

    if (myTX < 0)
	return false;
    return true;
}

bool
MOB::aiRandomWalk(bool orthoonly)
{
    int		dx, dy;
    int		angle, resultangle = 0;
    int		choice = 0;

    for (angle = 0; angle < 8; angle++)
    {
	rand_angletodir(angle, dx, dy);
	if (canMove(getX()+dx, getY()+dy))
	{
	    choice++;
	    if (!rand_choice(choice))
		resultangle = angle;
	}
	if (orthoonly)
	    angle++;
    }

    if (choice)
    {
	// Move in the direction
	rand_angletodir(resultangle, dx, dy);
	return actionWalk(dx, dy);
    }

    // Failed
    return false;
}

bool
MOB::aiCharge(MOB *foe, AI_NAMES aitype, bool orthoonly)
{
    if (aiAcquireTarget(foe))
    {
	if (aitype == AI_FLANK)
	{
	    if (aiFlankTo(myTX, myTY))
		return true;
	}
	else
	{
	    if (aiMoveTo(myTX, myTY, orthoonly))
		return true;
	}
    }

    return false;
}

bool
MOB::aiRangeAttack(MOB *target)
{
    // See if we have a ranged attack at all!
    if (!defn().range_valid)
	return false;

    int		 dx, dy;
    MOB		*a = MOB::getAvatar(), *v;

    // Override the target from the avatar...
    if (target)
	a = target;

    if (!a)
	return false;

    dx = SIGN(a->getX() - getX());
    dy = SIGN(a->getY() - getY());

    // If we are in melee range, don't use our ranged attack!
    if (MAX(ABS(a->getX()-getX()), ABS(a->getY()-getY())) <= 1)
    {
	return false;
    }

    // Try ranged attack.
    v = getMap()->traceBullet(getX(), getY(), getRangedRange(), dx, dy);
    if (v == a)
    {
	// Potential range attack.
	return actionFire(dx, dy);
    }

    return false;
}

bool
MOB::aiMoveTo(int x, int y, bool orthoonly)
{
    int		dx, dy, dist;
    int		angle, resultangle = 0, i;
    int		choice = 0;

    dx = SIGN(x - getX());
    dy = SIGN(y - getY());

    if (orthoonly && dx && dy)
    {
	// Ensure source is orthogonal.
	if (rand_choice(2))
	    dx = 0;
	else
	    dy = 0;
    }

    dist = MAX(ABS(x - getX()), ABS(y - getY()));

    if (dist == 1)
    {
	// Attack!
	if (canMove(x, y, false))
	    return actionBump(dx, dy);
	return false;
    }

    // Move in general direction, preferring a straight line.
    angle = rand_dirtoangle(dx, dy);

    for (i = 0; i <= 2; i++)
    {
	if (orthoonly && (i & 1))
	    continue;

	rand_angletodir(angle-i, dx, dy);
	if (canMove(getX()+dx, getY()+dy))
	{
	    choice++;
	    if (!rand_choice(choice))
		resultangle = angle-i;
	}
	
	rand_angletodir(angle+i, dx, dy);
	if (canMove(getX()+dx, getY()+dy))
	{
	    choice++;
	    if (!rand_choice(choice))
		resultangle = angle+i;
	}

	if (choice)
	{
	    rand_angletodir(resultangle, dx, dy);

	    // If wew can leap, maybe leap?
	    if (defn().canleap)
	    {
		if (dist > 2 &&
		    canMove(getX()+2*dx, getY()+2*dy))
		{
		    if (getMap()->isFOV(getX(), getY()))
		    {
			msg_format("%S <leap>.", this);
		    }
		    dx *= 2;
		    dy *= 2;
		}
	    }
	    
	    return actionWalk(dx, dy);
	}
    }

    return false;
}

bool
MOB::aiKillPathTo(MOB *target)
{
    if (!target)
	return false;
    
    return aiPathFindTo(target->getX(), target->getY());
}

bool
MOB::aiPathFindTo(int x, int y)
{
    int		dx, dy, ndx, ndy;
    int		curdist, dist, numvalid;
    bool	found;

    // see if already there.
    if (getX() == x && getY() == y)
	return false;

    curdist = getMap()->getDist(POS(x, y), getPos());
    
    // We don't care if this square is unreachable, as if it is the
    // square with the mob, almost by definition it is unreachable
    
    found = false;
    numvalid = 0;
    for (dy = -1; dy <= 1; dy++)
    {
	for (dx = -1; dx <= 1; dx++)
	{
	    // Ignore spots we can't move
	    if (!canMove(getX()+dx, getY()+dy, true))
		continue;
	    if (dx || dy)
	    {
		dist = getMap()->getDist(POS(x,y), POS(getX()+dx, getY()+dy));
		if (dist >= 0 && dist < curdist)
		{
		    found = true;
		    ndx = dx;
		    ndy = dy;
		    curdist = dist;
		    numvalid = 1;
		}
		else if (dist == curdist)
		{
		    // We have an option of closest, pick at random.
		    numvalid++;
		    if (!rand_choice(numvalid))
		    {
			ndx = dx;
			ndy = dy;
		    }
		}
	    }
	}
    }

    // If we didn't find a direction, abort
    if (!found)
    {
	// No need to spam since it is likely just someone in our way.
	// msg_format("%S cannot find a path there.", this);
	return false;
    }

    // Otherwise, act.
    return actionWalk(ndx, ndy);
}

bool
MOB::aiFlankTo(int x, int y)
{
    int		dx, dy, dist;
    int		angle, resultangle = 0, i, j;
    int		choice = 0;

    dx = SIGN(x - getX());
    dy = SIGN(y - getY());

    dist = MAX(ABS(x - getX()), ABS(y - getY()));

    if (dist == 1)
    {
	// Attack!
	if (canMove(x, y, false))
	    return actionBump(dx, dy);
	return false;
    }

    // Move in general direction, preferring a straight line.
    angle = rand_dirtoangle(dx, dy);

    for (j = 0; j <= 2; j++)
    {
	// To flank, we prefer the non-straight approach.
	switch (j)
	{
	    case 0:
		i = 1;
		break;
	    case 1:
		i = 0;
		break;
	    case 2:
		i = 2;
		break;
	}
	rand_angletodir(angle-i, dx, dy);
	if (canMove(getX()+dx, getY()+dy))
	{
	    choice++;
	    if (!rand_choice(choice))
		resultangle = angle-i;
	}
	
	rand_angletodir(angle+i, dx, dy);
	if (canMove(getX()+dx, getY()+dy))
	{
	    choice++;
	    if (!rand_choice(choice))
		resultangle = angle+i;
	}

	if (choice)
	{
	    rand_angletodir(resultangle, dx, dy);
	    return actionWalk(dx, dy);
	}
    }

    return false;
}

MOB *
MOB::aiFindFavoriteChild() const
{
    MOB		*favorite = 0, *mob;

    for (mob = getMap()->getMobHead(); mob; mob = mob->getNext())
    {
	if (!mob->defn().ischild)
	    continue;
	if (!favorite || favorite->myMind.myAttention < mob->myMind.myAttention)
	{
	    favorite = mob;
	}
    }

    return favorite;
}

ACTIVITY_NAMES
MOB::getActivity() const
{
    return (ACTIVITY_NAMES) glb_ai_statedefs[myAIState].activity;
}

AI_STATE_NAMES
MOB::aiChildSeekAttention()
{
    MOB		*favorite;
    int		 delta;

    // Determine the favorite child...
    favorite = aiFindFavoriteChild();

    if (!favorite || favorite == this)
    {
	// YEah!  I'm winning the game.  Keep up the good work.
	// Exception if we are idle.
	if (myAIState == AI_STATE_NONE)
	{
	    return (AI_STATE_NAMES) glb_activitydefs[myMind.favoriteActivity()].state;
	}
	return myAIState;
    }

    delta = favorite->myMind.myAttention - myMind.myAttention;

    // Choose between my favorite activity and the favorite's child's activity
    if (rand_choice(myMind.myAttention) > rand_choice(favorite->myMind.myAttention))
	favorite = this;

    ACTIVITY_NAMES activity = favorite->myMind.favoriteActivity();

    // Include into our gradient how much more effective our most effective
    // activity was compared to our current activity.
    delta += MAX(myMind.myActivity[getActivity()]-myMind.myActivity[activity],0);

    // If the gradient is big enough, we will switch tasks.
    if (myAIState == AI_STATE_NONE || !rand_choice(100 / (delta+1)))
    {
	return (AI_STATE_NAMES) glb_activitydefs[activity].state;
    }

    // Status quo it is!
    return myAIState;
}

bool
MOB::aiChild()
{
    // Bump out of default explore.
    if (myAIState == AI_STATE_NONE)
    {
	myAIState = aiChildSeekAttention();
    }

    // Watch the avatar...
    MOB		*avatar = getAvatar();
    
    if (avatar)
    {
	// See how dangerous they are playing.
	if (getMap()->isFOV(avatar->getPos()))
	{
	    int		depth;
	    depth = getMap()->waterDepth(avatar->getPos());
	    if (depth < myMind.myWaterDist)
		myMind.myWaterDist = depth;
	}
    }
    
    // If we are playing, reduce our playcount.
    if (myPlayCount)
    {
	myPlayCount--;

	// Stop playing if we hit the end.
	if (!myPlayCount)
	{
	    if (myAIState == AI_STATE_TAGIT ||
		myAIState == AI_STATE_TAGESCAPE)
	    {
		// bored of play, return to main.
		myAIState = AI_STATE_NONE;
		myAIState = aiChildSeekAttention();
	    }
	}
    }
    else
    {
	// What do I want to do now?
	// Not playing, chance to decide to start.
	if (!rand_choice(100))
	{
	    msg_format("%S <grow> bored.", this);
	    clearBoredom();
    
	    myAIState = AI_STATE_TAGESCAPE;
	    myPlayCount = rand_choice(50)+20;
	}
	else
	{
	    myAIState = aiChildSeekAttention();
	}
    }
    
    // Determine our current state...
    switch (myAIState)
    {
	case AI_STATE_PILEROCKS:
	    return aiChildPileRocks();
	case AI_STATE_TAGIT:
	    return aiChildTagIt();
	case AI_STATE_TAGESCAPE:
	    return aiChildTagNotIt();
	case AI_STATE_SULK:
	    return aiChildSulk();
	case AI_STATE_RECALL:
	    return aiChildRecall();
    }

    return aiRandomWalk();
}

bool
MOB::aiChildTagIt()
{
    MOB		*target;

    if (!rand_choice(15))
	myMind.myTagger = MOB_NONE;
    
    // I'm it.  Look for a nearby child to tag and tag them.
    target = aiFindCloseChild(myMind.myTagger);

    // If there is no child, wander randomly
    if (!target)
    {
	myPlayCount = 0;		// end game
	return aiRandomWalk();
    }

    // Small chance of giving up on chasing this guy.
    if (!rand_choice(25))
	myMind.myTagger = target->getDefinition();

    // If we are carrying anything, drop it.
    if (myInventory)
    {
	return actionDrop();
    }
    
    // If we are adjacent, we can tag them!
    if (MAP::dist(this, target) == 1)
    {
	msg_format("%S <yell>: \"Tag!  You're it!\"", this);
	target->myAIState = AI_STATE_TAGIT;
	target->myMind.myTagger = getDefinition();
	// The target joins in on the game
	// To avoid the game going forever, we inherit the playcount.
	target->myPlayCount = myPlayCount;
	myAIState = AI_STATE_TAGESCAPE;

	// We flee immediately to avoid tag-backs.
	return aiFleeFrom(target->getX(), target->getY());
    }

    // Chase...
    if (!aiPathFindTo(target->getX(), target->getY()))
	return aiRandomWalk();
    return true;
}

bool
MOB::aiChildTagNotIt()
{
    MOB		*it;
    
    it = aiFindIt();
    if (!it)
    {
	// No one is it.  So, it shall be me!
	myAIState = AI_STATE_TAGIT;
	msg_format("%S <yell>: \"Ready or not, here I come!\"", this);
	return aiChildTagIt();
    }

    // If we are carrying anything, drop it.
    if (myInventory)
    {
	return actionDrop();
    }
    
    // Check our distance to it.  It isn't fun to get too far away in
    // tag, want to stand and taunt occasionally :>
    if (getMap()->getDist(it->getPos(), getPos()) > 4)
    {
	if (!rand_choice(30))
	{
	    msg_format("%S <yell>: \"Nyah, Nyah, you can't catch me!\"", this);
	}
	if (!aiPathFindTo(it->getPos()))
	    return aiRandomWalk();
	return true;
    }
    // We flee the it.
    return aiFleeFrom(it->getX(), it->getY());
}

bool
MOB::aiChildPileRocks()
{
    // If not carrying a rock, find one!
    if (!myInventory)
    {
	ITEM		*rock;
	rock = aiFindCloseRock();
	if (!rock)
	{
	    // Switch back to tag.
	    msg_format("%S <yell>: \"All done!\"", this);
	    myAIState = AI_STATE_TAGESCAPE;
	    return false;
	}

	// If rock is here, pick it up (AI is responsible for avoiding
	// double pickups)
	if (rock->getX() == getX() && rock->getY() == getY())
	{
	    // CLear our drop locatoin.
	    myTX = -1;
	    return actionPickup();
	}

	// Path find to the rock.
	if (!aiPathFindTo(rock->getX(), rock->getY()))
	{
	    // Failed to find a path, likely blocked
	    return aiRandomWalk();
	}

	return true;
    }

    // Carrying a rock... See if our target location is still reachable.
    if (myTX < 0 || !getMap()->isConnected(getPos(), POS(myTX, myTY), myMind.myWaterDist))
    {
	aiChooseRockDrop(myTX, myTY);
    }

    // Are we at good place to drop?
    if (getX() == myTX && getY() == myTY)
    {
	// We chose to drop for a good reason.
	return actionDrop();
    }
    
    // Move to it...
    if (!aiPathFindTo(myTX, myTY))
    {
	// Our path is blocked...
	// Consider just dropping here.
	if (aiIsRockPosGood(getPos()))
	    return actionDrop();

	return false;
    }
    return true;
}

bool
MOB::aiChildSulk()
{
    // Random chance to end sulk,
    if (!rand_choice(20 * (myMind.myScolds+1)))
    {
	myAIState = AI_STATE_NONE;
	return aiChild();
    }
    // Wander off in a fit.
    return aiRandomWalk();
}

bool
MOB::aiChildRecall()
{
    MOB		*dad = MOB::getAvatar();

    if (!dad)
    {
	myAIState = AI_STATE_NONE;
	return false;
    }

    // Check to see if we are close enough
    if (getMap()->getDist(dad->getPos(), getPos()) <= 2)
    {
	myAIState = AI_STATE_NONE;
	return false;
    }

    // Path find to the dad
    if (!aiPathFindTo(dad->getPos()))
    {
	if (!rand_choice(5))
	    msg_format("%S cannot find a safe way back.", this);
	return aiRandomWalk();
    }
    return true;
}

bool
MOB::aiReceivePraise()
{
    myMind.payAttention(getActivity(), 1);
    
    return true;
}

bool
MOB::aiReceiveRecall()
{
    myMind.payAttention(getActivity(), 1);

    switch (getActivity())
    {
	case ACTIVITY_TAG:
	case ACTIVITY_SULK:
	    // Treat this as a scold as we don't want to recall.
	    myMind.myScolds++;
	    break;
    }
    myAIState = AI_STATE_RECALL;
    
    return true;
}

bool
MOB::aiReceiveScold()
{
    myMind.payAttention(getActivity(), 1);
    // This has the effect of increasing the total attention that
    // negative reinforcment provides and ensuring sulking gets
    // some interest.
    myMind.payAttention(ACTIVITY_SULK, 1);
    myMind.myScolds++;

    // Reaction depends on what we think we did wrong...
    switch (getActivity())
    {
	case ACTIVITY_TAG:
	    // Stop playing.
	    myAIState = AI_STATE_PILEROCKS;
	    myPlayCount = 0;
	    break;

	case ACTIVITY_PILEROCKS:
	    // Must have done something wrong in piling the rocks.
	    // If I have a rock, drop it and mark it final.
	    if (myInventory)
	    {
		myInventory->setFinalPos(true);
		actionDrop();
	    }
	    else
	    {
		// If there is a rock here, I must have dropped it in
		// the wrong spot.
		if (getMap()->getItem(getX(), getY()))
		    actionPickup();
		else
		{
		    // Dad just hates me.
		    myAIState = AI_STATE_SULK;
		}
	    }
	    break;

	case ACTIVITY_RECALL:
	    // But I *am* recalling!
	    myAIState = AI_STATE_SULK;
	    break;

	case ACTIVITY_SULK:
	    // Obviously, not suppsoed to be sulking.
	    myAIState = AI_STATE_PILEROCKS;
	    break;
    }

    return true;
}
