/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        item.cpp ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#include "item.h"

#include <assert.h>

#include "map.h"

#include <iostream>
using namespace std;

//
// Item Definitions
//

ITEM::ITEM()
{
    myX = myY = -1;
    myDefinition = (ITEM_NAMES) 0;
    myNext = 0;
    myCount = 1;
    myTimer = -1;
    myFinalPos = false;
}

ITEM::~ITEM()
{
    delete myNext;
}

ITEM *
ITEM::copy() const
{
    ITEM *item;

    item = new ITEM();

    *item = *this;

    // Undo entirely personal things.
    item->myNext = 0;
    return item;
}

ITEM *
ITEM::create(ITEM_NAMES item, int depth)
{
    ITEM	*i;

    assert(item >= ITEM_NONE && item < NUM_ITEMS);

    i = new ITEM();

    i->myDefinition = item;
    i->myTimer = glb_itemdefs[item].timer;

    return i;
}

ITEM *
ITEM::createRandom(int depth)
{
    ITEM_NAMES		item;

    item = itemFromHash(rand_choice(65536*32767));
    return create(item, depth);
}

ITEM_NAMES
ITEM::itemFromHash(unsigned int hash)
{
    int		totalrarity, rarity;
    int		i;

    // Find total rarity of items
    totalrarity = 0;
    for (i = ITEM_NONE; i < NUM_ITEMS; i++)
    {
	rarity = defn((ITEM_NAMES)i).rarity;
	totalrarity += rarity;
    }

    hash %= totalrarity;
    for (i = ITEM_NONE; i < NUM_ITEMS; i++)
    {
	rarity = defn((ITEM_NAMES)i).rarity;

	// We must do >= or a hash value of 0 will break for
	// ITEM_NONE creating programming errors.
	if (hash >= (unsigned) rarity)
	    hash -= rarity;
	else
	    break;
    }
    return (ITEM_NAMES) i;
}

VERB_PERSON
ITEM::getPerson() const
{
    return VERB_IT;
}

BUF
ITEM::getName(bool usearticle) const
{
    BUF		buf;

    if (getTimer() >= 0)
    {
	// A timed item...
	buf.sprintf("%s (%d)", defn().name, getTimer());
    }
    else
    {
	// A normal item.
	buf.reference(defn().name);
    }

    return gram_createcount(buf, myCount, usearticle);
}

const char *
ITEM::getDetailedDescription() const
{
    switch (getDefinition())
    {
	default:
	    return 0;
    }
}

void
ITEM::getLook(u8 &symbol, ATTR_NAMES &attr) const
{
    symbol = defn().symbol;
    attr = (ATTR_NAMES) defn().attr;
}

void
ITEM::move(MAP *map, int x, int y)
{
    myX = x;
    myY = y;
    if (map)
	map->setFlag(getPos(), MAPFLAG_ITEM, true);
}

bool
ITEM::canStackWith(const ITEM *stack) const
{
    if (getDefinition() != stack->getDefinition())
	return false;

    // Disable stacking of weapons/books/armour as we want
    // to be able to delete them easily
    if (defn().unstackable)
	return false;

    // No reason why not...
    return true;
}

void
ITEM::combineItem(const ITEM *item)
{
    // Untimed items stack.
    if (getTimer() < 0)
	myCount += item->getStackCount();
    // Timed items add up charges.
    if (getTimer() >= 0)
    {
	assert(item->getTimer() >= 0);
	myTimer += item->getTimer();
    }

    assert(myCount >= 0);
    if (myCount < 0)
	myCount = 0;
}

ITEM *
ITEM::splitStack(int newstack)
{
    assert(getStackCount() > newstack);

    ITEM		*result;

    result = copy();

    result->myCount = newstack;
    myCount -= newstack;

    return result;
}

bool
ITEM::runHeartbeat()
{
    if (myTimer >= 0)
    {
	if (!myTimer)
	    return true;
	myTimer--;
    }
    return false;
}

void
ITEM::save(ostream &os) const
{
    int		val;

    val = myDefinition;
    os.write((const char *) &val, sizeof(int));

    os.write((const char *) &myX, sizeof(int));
    os.write((const char *) &myY, sizeof(int));

    os.write((const char *) &myCount, sizeof(int));
    os.write((const char *) &myTimer, sizeof(int));
    os.write((const char *) &myFinalPos, sizeof(bool));
}

ITEM *
ITEM::load(istream &is)
{
    int		val;
    ITEM	*i;

    i = new ITEM();

    is.read((char *)&val, sizeof(int));
    i->myDefinition = (ITEM_NAMES) val;

    is.read((char *)&i->myX, sizeof(int));
    is.read((char *)&i->myY, sizeof(int));

    is.read((char *)&i->myCount, sizeof(int));
    is.read((char *)&i->myTimer, sizeof(int));
    is.read((char *)&i->myFinalPos, sizeof(bool));

    return i;
}

void
ITEM::getWeaponStats(int &power, int &accuracy, int &consistency) const
{
    power = accuracy = consistency = 0;

    return;
}

DPDF
ITEM::getMeleeDPDF() const
{
    return DPDF(0);
}

void
ITEM::getRangeStats(int &range, int &power, int &consistency, int &area, int &mana) const
{
    range = power = consistency = mana = 0;
    return;
}

int
ITEM::getRangeRange() const
{
    // Compute our power, accuracy, and consistency.
    int		range, d;

    getRangeStats(range, d, d, d, d);

    return range;
}

int
ITEM::getRangeMana() const
{
    // Compute our power, accuracy, and consistency.
    int		mana, d;

    getRangeStats(d, d, d, d, mana);

    return mana;
}

int
ITEM::getRangeArea() const
{
    // Compute our power, accuracy, and consistency.
    int		area, d;

    getRangeStats(d, d, d, area, d);

    return area;
}

DPDF
ITEM::getRangeDPDF() const
{
    return DPDF(0);
}
