/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        map.cpp ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#include "map.h"
#include "gfxengine.h"
#include "mob.h"
#include "msg.h"
#include "item.h"
#include "text.h"
#include "speed.h"

#include <assert.h>

#include <iostream>
#include <fstream>
using namespace std;

#define CODE_TOP 0
#define CODE_RIGHT 1
#define CODE_BOT 2
#define CODE_LEFT 3

#define SOLID_CHAR ((int)'^')
#define SOLID_CODE ((((SOLID_CHAR << 8) | SOLID_CHAR) << 8) | SOLID_CHAR)

#define FORALL_MAP(x, y) \
    for (y = 0; y < getHeight(); y++) \
	for (x = 0; x < getWidth(); x++)

#define FORALL_MAPPOS(p) \
    for (p.y = 0; p.y < getHeight(); p.y++) \
	for (p.x = 0; p.x < getWidth(); p.x++)

// Class systems.
int
findClass(MAPDATA &connect, int x, int y)
{
    int		idx, fclass, newclass, nx, ny;

    fclass = connect(x, y);
    
    if (fclass == -1)
	return -1;

    idx = x + y * 65536;

    if (fclass == idx)
	return fclass;

    nx = fclass & 65535;
    ny = fclass >> 16;
    newclass = findClass(connect, nx, ny);
    if (newclass != fclass)
	connect.data(x, y) = newclass;
    return newclass;
}

int
findFromClass(MAPDATA &connect, int idx)
{
    int		x, y;

    x = idx & 65535;
    y = idx >> 16;
    return findClass(connect, x, y);
}

void
connectClass(MAPDATA &connect, int x, int y, int nx, int ny)
{
    int		fclass, baseclass;

    baseclass = findClass(connect, x, y);
    fclass = findClass(connect, nx, ny);
    if (fclass != baseclass)
    {
	nx = fclass & 65535;
	ny = fclass >> 16;
	connect.data(nx, ny) = baseclass;
    }
}

void
connectToClass(MAPDATA &connect, int x, int y, int newclass)
{
    int nx, ny;

    nx = newclass & 65535;
    ny = newclass >> 16;

    connectClass(connect, nx, ny, x, y);
}

//
// Stores the legend information
//
class LEGEND
{
public:
	     LEGEND() {}
	     LEGEND(u8 symbol, const char *descr);
	     LEGEND(u8 symbol, const char *descr,
		     TILE_NAMES tile, MOB_NAMES mob, ITEM_NAMES item,
		     int itemcount);
	     LEGEND(int blah) {}
	     LEGEND(const LEGEND &l) { *this = l; }
    virtual ~LEGEND();

    const LEGEND &operator=(const LEGEND &s)
    {
	mySymbol = s.mySymbol;
	myDescr = s.myDescr;
	myMob = s.myMob;
	myItem = s.myItem;
	myTile = s.myTile;
	myItemCount = s.myItemCount;
	return *this;
    }

    const char	*getDescr() const { return myDescr.buffer(); }
    u8		 getSymbol() const { return mySymbol; }

protected:
    u8		 mySymbol;
    BUF		 myDescr;

public:
    // These are used for saving.
    MOB_NAMES	 myMob;
    ITEM_NAMES	 myItem;
    int		 myItemCount;
    TILE_NAMES	 myTile;
};

LEGEND::LEGEND(u8 symbol, const char *descr)
{
    // Strip leading spaces.
    while (ISSPACE(*descr))
	descr++;

    mySymbol = symbol;
    myDescr.strcpy(descr);
}

LEGEND::LEGEND(u8 symbol, const char *descr, TILE_NAMES tile, MOB_NAMES mob, ITEM_NAMES item, int itemcount)
{
    // Strip leading spaces.
    while (ISSPACE(*descr))
	descr++;

    mySymbol = symbol;
    myDescr.strcpy(descr);

    myTile = tile;
    myMob = mob;
    myItem = item;
    myItemCount = itemcount;
}

LEGEND::~LEGEND()
{
}

//
// Each of these is a 5x5 tile piece.
//
class PIECE
{
public:
    PIECE(u8 map[5][5], int sym);
    ~PIECE();

    bool	doesMatch(int top, int right, int bot, int left) const;
    int		getCode(int side) const;
    u8		codeSet(int x, int y) const;

    void	writeToMap(MAP *map, const LEGENDLIST &legend, int mx, int my) const;

    PIECE	*getNext() const { return myNext; }
    void	setNext(PIECE *piece) { myNext = piece; }
    
protected:
    u8		myMap[5][5];

    PIECE	*myNext;
};

PIECE::PIECE(u8 map[5][5], int sym)
{
    int		x, y, rx, ry, t;

    for (x = 0; x < 5; x++)
    {
	for (y = 0; y < 5; y++)
	{
	    if (sym & 1)
		rx = x;
	    else
		rx = 4 - x;
	    if (sym & 2)
		ry = y;
	    else
		ry = 4 - y;
	    if (sym & 4)
	    {
		t = rx;
		rx = ry;
		ry = t;
	    }

	    myMap[rx][ry] = map[x][y];
	}
    }
	
    myNext = 0;
}

PIECE::~PIECE()
{
    delete myNext;
}

bool
PIECE::doesMatch(int top, int right, int bot, int left) const
{
    if (top != -1 && getCode(CODE_TOP) != top)
	return false;
    if (right != -1 && getCode(CODE_RIGHT) != right)
	return false;
    if (bot != -1 && getCode(CODE_BOT) != bot)
	return false;
    if (left != -1 && getCode(CODE_LEFT) != left)
	return false;

    return true;
}

int
PIECE::getCode(int side) const
{
    int		code = 0;
    int		x, y;

    // Because we care not about corners, our code is merely
    // our 3 central squares appended together.

    switch (side)
    {
	case CODE_TOP:
	    y = 0;
	    for (x = 1; x < 4; x++)
	    {
		code <<= 8;
		code |= codeSet(x, y);
	    }
	    break;
	case CODE_RIGHT:
	    x = 4;
	    for (y = 1; y < 4; y++)
	    {
		code <<= 8;
		code |= codeSet(x, y);
	    }
	    break;
	case CODE_BOT:
	    y = 4;
	    for (x = 1; x < 4; x++)
	    {
		code <<= 8;
		code |= codeSet(x, y);
	    }
	    break;
	case CODE_LEFT:
	    x = 0;
	    for (y = 1; y < 4; y++)
	    {
		code <<= 8;
		code |= codeSet(x, y);
	    }
	    break;
	default:
	    assert(0);
	    break;
    }
    return code;
}

u8
PIECE::codeSet(int x, int y) const
{
    return myMap[y][x];
}

void
PIECE::writeToMap(MAP *map, const LEGENDLIST &legend, int mx, int my) const
{
    int		x, y;

    for (y = 0; y < 5; y++)
    {
	for (x = 0; x < 5; x++)
	{
	    map->parseLegend(mx+x, my+y, legend, myMap[y][x]);
	}
    }
}

class PIECEDICT
{
public:
    PIECEDICT();
    ~PIECEDICT();
    
    PIECE	*getRandomPiece(int top, int right, int bot, int left);
    void	 append(PIECE *piece);
    const LEGENDLIST &legend() const { return myLegend; }

protected:
    PIECE	*myPieces;
    LEGENDLIST myLegend;
};

PIECEDICT::PIECEDICT()
{
    char		buf[100];
    ifstream		is("../rooms/piecelist.dict");
    u8			map[5][5];
    int			x, y, i;

    myPieces = 0;
    myLegend.clear();

    while (is.getline(buf, 100))
    {
	text_striplf(buf);
	// Comments.
	if (buf[0] == '#' && buf[1] == ' ')
	    continue;

	// Legends:
	if (buf[0] && buf[1] == ':')
	{
	    myLegend.append(LEGEND(buf[0], &buf[2]));
	}

	if (buf[0] == '+')
	{
	    y = 0;
	    while (is.getline(buf, 100))
	    {
		text_striplf(buf);
		if (buf[0] == '+')
		    break;
		if (y < 5 && buf[0] == '|')
		{
		    for (x = 0; x < 5; x++)
		    {
			if (!buf[x+1])
			    break;
			map[y][x] = buf[x+1];
		    }
		    y++;
		}
	    }

	    // Got a new map.  Build the pieces.
	    for (i = 0; i < 8; i++)
	    {
		append(new PIECE(map, i));
	    }
	}
    }
}

PIECEDICT::~PIECEDICT()
{
    delete myPieces;
}

PIECE *
PIECEDICT::getRandomPiece(int top, int right, int bot, int left)
{
    int		 nummatch = 0;
    PIECE	*cur = 0, *match = 0;
    
    for (cur = myPieces; cur; cur = cur->getNext())
    {
	if (cur->doesMatch(top, right, bot, left))
	{
	    nummatch++;

	    if (!rand_choice(nummatch))
		match = cur;
	}	
    }

    return match;
}

void
PIECEDICT::append(PIECE *piece)
{
    piece->setNext(myPieces);
    myPieces = piece;
}

//
// MAP implementation
//
MAP::MAP()
{
    MOB_NAMES mob;
    
    myWidth = myHeight = 0;
    myMobs = 0;
    myItems = 0;

    myDownMap = myUpMap = 0;
    myMobSafeNext = 0;
    myTotalRooms = 0;
    myRooms = 0;

    FOREACH_MOB(mob)
    {
	myEverHad[mob] = false;
    }

    // Ensure our default lastx/lasty are reasonable.
    int		sx, sy, cx, cy, w, h;

    // Stolen from our basic redraw.
    sx = sy = 1;
    w = 35;
    h = 25;
    cx = 17;
    cy = 12;
    
    cx = cx - w / 2;
    cy = cy - h / 2;

    myLastX = sx;
    myLastY = sy;
    myLastMX = cx;
    myLastMY = cy;
    myLastW = w;
    myLastH = h;

    clearAllCaches();
}

MAP::~MAP()
{
    delete myMobs;
    delete myItems;

    // Delete any up and down linked maps.
    if (myDownMap)
    {
	myDownMap->myUpMap = 0;
	delete myDownMap;
    }
    
    if (myUpMap)
    {
	myUpMap->myDownMap = 0;
	delete myUpMap;
    }
}

int
MAP::dist(int x1, int y1, int x2, int y2)
{
    int		dx, dy;

    dx = x2 - x1;
    dy = y2 - y1;

    return MAX(ABS(dx), ABS(dy));
}

int
MAP::dist(const MOB *a, const MOB *b)
{
    // Less typing, more function calls :>
    return dist(a, b->getX(), b->getY());
}

int
MAP::dist(const MOB *a, int x, int y)
{
    return dist(a->getX(), a->getY(), x, y);
}

bool
MAP::isLinedUp(int x1, int y1, int x2, int y2)
{
    int		dx, dy;

    dx = x2 - x1;
    dy = y2 - y1;

    // Check ortho...
    if (!dx || !dy)
	return true;

    // Check diagonal.
    if (ABS(dx) == ABS(dy))
	return true;

    return false;
}

bool
MAP::isLinedUp(const MOB *a, const MOB *b)
{
    return isLinedUp(a, b->getX(), b->getY());
}

bool
MAP::isLinedUp(const MOB *a, int x, int y)
{
    return isLinedUp(a->getX(), a->getY(), x, y);
}

MAP *
MAP::copyAll() const
{
    MAP		*map, *peek, *dup, *lastdup;

    map = copy();
    lastdup = map;
    // Copy upwards.
    for (peek = peekUpMap(); peek; peek = peek->peekUpMap())
    {
	dup = peek->copy();
	dup->setDownMap(lastdup);
	lastdup->setUpMap(dup);
	lastdup = dup;
    }

    // Copy downwards
    lastdup = map;
    for (peek = peekDownMap(); peek; peek = peek->peekDownMap())
    {
	dup = peek->copy();
	dup->setUpMap(lastdup);
	lastdup->setDownMap(dup);
	lastdup = dup;
    }

    // REturn ourselves.
    return map;
}

MAP *
MAP::copy() const
{
    MAP		*map;

    map = new MAP();
    *map = *this;

    map->myUpMap = 0;
    map->myDownMap = 0;

    // Copy the mobs, being careful to keep the same order.
    MOB		*mob, *dupmob, *lastmob;

    map->myMobs = 0;
    map->myMobSafeNext = 0;
    lastmob = 0;
    for (mob = getMobHead(); mob; mob = mob->getNext())
    {
	dupmob = mob->copy();
	dupmob->setMap(map);
	if (lastmob)
	    lastmob->setNext(dupmob);
	else
	    map->myMobs = dupmob;
	lastmob = dupmob;
    }

    // Ditto for items on the ground.
    ITEM	*item, *dup, *last = 0;

    map->myItems = 0;
    for (item = myItems; item; item = item->getNext())
    {
	dup = item->copy();
	if (last)
	    last->setNext(dup);
	else
	    map->myItems = dup;
	last = dup;
    }

    return map;
}

void
MAP::locateAvatar(MOB_NAMES avatarname)
{
    MAP		*map;
    
    MOB::setAvatar(0);
    
    for (map = this; map; map = map->peekDownMap())
    {
	if (MOB::getAvatar())
	    break;
	map->locateAvatarLocal(avatarname);
    }

    for (map = peekUpMap(); map; map = map->peekUpMap())
    {
	if (MOB::getAvatar())
	    break;
	map->locateAvatarLocal(avatarname);
    }
}

void
MAP::locateAvatarLocal(MOB_NAMES avatarname)
{
    MOB		*mob;
    for (mob = getMobHead(); mob; mob = mob->getNext())
    {
	if (mob->getDefinition() == avatarname)
	{
	    MOB::setAvatar(mob);
	    break;
	}
    }
}

void
MAP::setMapName(const char *name)
{
    myMapName.strcpy(name);
    myMapName.stripExtension();
}

void
MAP::loadMap(const char *fname)
{
    char		line[1002];

    ifstream		is(fname);
    LEGENDLIST		legendlist;
    PTRLIST<char *>	linelist;
    int			w = 0, h = 0;
    int			x, y;
    u8			symbol;

    myOrigState.clear();

    while (is.getline(line, 1000))
    {
	text_striplf(line);

	myOrigState.strcat(line);
	myOrigState.append('\n');

	// Check for special keywords.
	if (!strncmp(line, "Intro:", strlen("Intro:")))
	{
	    myIntro.clear();
	    while (is.getline(line, 1000))
	    {
		text_striplf(line);

		myOrigState.strcat(line);
		myOrigState.append('\n');

		if (!line[0] || ISSPACE(line[0]))
		{
		    int firstnonws;

		    for (firstnonws = 0; line[firstnonws] && ISSPACE(line[firstnonws]); firstnonws++);
		    if (!line[firstnonws])
		    {
			myIntro.strcat("\n\n");
		    }
		    else
		    {
			// If indented, we always indent.
			if (firstnonws > 4)
			{
			    myIntro.strcat("\n");
			    firstnonws = 4;
			}

			// If previous paragraph ended (\n)
			// or ended a sentence we want to add appropriate
			// spaces
			if (myIntro.isstring() && myIntro.lastchar() != '\n')
			{
			    if (gram_isendsentence(myIntro.lastchar()))
				myIntro.append(' ');
			    myIntro.append(' ');
			}

			myIntro.strcat(&line[firstnonws]);
		    }
		}
		else
		{
		    // A real line again, consume!
		    break;
		}
	    }
	}

	// Comments.
	if (line[0] == '#' && line[1] == ' ')
	    continue;

	// Legends:
	if (line[0] && line[1] == ':')
	{
	    legendlist.append(LEGEND(line[0], &line[2]));
	}

	if (line[0] == '+' && line[1] == '-')
	{
	    // We have start of map proper.
	    for (w = 1; line[w] == '-'; w++);

	    if (line[w] != '+')
	    {
		cerr << "Map does not have proper bounded map." << endl;
	    }

	    w--;

	    // Load the actual map lines.
	    while (is.getline(line, 1000))
	    {
		text_striplf(line);
		myOrigState.strcat(line);
		myOrigState.append('\n');

		if (line[0] == '+')
		    break;

		if (line[0] != '|')
		{
		    cerr << "Map line missing | prefix." << endl;
		}
		if (line[w+1] != '|')
		{
		    cerr << "Map line too short!" << endl;
		}
		line[w+1] = '\0';
		linelist.append(strdup(&line[1]));
	    }

	    h = linelist.entries();
	}
    }

    if (!w)
    {
	// No map found, abort.
	cerr << "No map segment found!" << endl;
	gfx_shutdown();
	exit(0);
    }

    // Build our map.
    myWidth = w;
    myHeight = h;

    myTiles = MAPDATA(myWidth, myHeight);
    myTileFlags = MAPDATA(myWidth, myHeight);
    myWetMap = MAPDATA(myWidth, myHeight);
    myTiles.constant(TILE_INVALID);
    myTileFlags.constant(0);
    myWetMap.constant(0);

    FORALL_MAP(x, y)
    {
	symbol = linelist(y)[x];

	parseLegend(x, y, legendlist, symbol);
    }

    // Delete line list.
    for (y = 0; y < myHeight; y++)
    {
	free(linelist(y));
    }

    clearAllCaches();
}

void
MAP::parseLegend(int x, int y, const LEGENDLIST &list,
		u8 symbol)
{
    // Find the legend.
    int			 i, j;
    const char		*s, *comma;
    char		*token;
    bool		 foundsymbol = false;

    for (i = 0; i < list.entries(); i++)
    {
	if (list(i).getSymbol() == symbol)
	{
	    s = list(i).getDescr();

	    foundsymbol = true;
	    
	    while (*s)
	    {
		for (comma = s; *comma && *comma != ','; comma++)
		{
		    if (*comma == '(')
		    {
			while (*comma && *comma != ')')
			{
			    comma++;
			}
		    }
		}
	    
		token = strdup(s);
		token[comma - s] = '\0';

		// We wish to apply token to this dude.
		if (token[0] == '(')
		{
		    // We have inheritance!
		    if (token[1] == symbol)
		    {
			cerr << "Naive infinite recursion: Try harder!" << endl;
			return;
		    }
		    parseLegend(x, y, list, token[1]);
		}
		else
		{
		    bool	madewish = false;
		    // Run through our wish engine...
		    for (j = 0; j < NUM_TILES; j++)
		    {
			if (!strcmp(glb_tiledefs[j].legend, token))
			{
			    madewish = true;
			    setTile(x, y, (TILE_NAMES) j);
			}
		    }

		    for (j = 0; j < NUM_MOBS; j++)
		    {
			if (!strcmp(glb_mobdefs[j].name, token))
			{
			    madewish = true;
			    MOB		*mob;

			    mob = MOB::create((MOB_NAMES) j);
			    mob->move(x, y);
			    addMob(mob);
			}
		    }

		    for (j = 0; j < NUM_ITEMS; j++)
		    {
			if (!strcmp(glb_itemdefs[j].name, token))
			{
			    madewish = true;
			    ITEM	*item;

			    item = ITEM::create((ITEM_NAMES) j);
			    item->move(0, x, y);
			    addItem(item);
			}
		    }

		    assert(madewish);
		}

		free(token);

		// Find next word.
		s = comma;
		while (*s == ',' || ISSPACE(*s))
		    s++;
	    }
	}
    }

    if (!foundsymbol)
    {
	cerr << "Did not find symbol " << symbol << endl;
    }
}

MAP *
MAP::getDownMap() const
{
    if (!myDownMap)
    {
#if 0
	myDownMap = new MAP();
	myDownMap->myUpMap = (MAP *) this;
	myDownMap->build(myTotalRooms);
#endif
    }
    return myDownMap;
}

MAP *
MAP::getUpMap() const
{
    if (!myUpMap)
    {
#if 0
	myUpMap = new MAP();
	myUpMap->myDownMap = (MAP *) this;
	myUpMap->build(1);
#endif
    }
    return myUpMap;
}

void
MAP::setDownMap(MAP *down)
{
    assert(!myDownMap);

    myDownMap = down;
}

void
MAP::setUpMap(MAP *up)
{
    assert(!myUpMap);

    myUpMap = up;
}

bool
MAP::findTile(TILE_NAMES tile, int &mx, int &my, bool avoidmob) const
{
    int		nummatch = 0;
    int		x, y;
    
    FORALL_MAP(x, y)
    {
	if (getTile(x, y) == tile)
	{
	    if (avoidmob && getMob(x, y))
		continue;

	    nummatch++;
	    if (!rand_choice(nummatch))
	    {
		mx = x;
		my = y;
	    }
	}
    }

    if (nummatch)
	return true;
    return false;
}
    
bool
MAP::findMoveable(int &mx, int &my, bool avoidmob) const
{
    int		nummatch = 0, i, trials;
    int		x, y;

    if (myMoveablePos.entries() == 0)
    {
	FORALL_MAP(x, y)
	{
	    if (isPassable(x, y))
	    {
		myMoveablePos.append(POS(x, y));

	    }
	}
    }
    if (!myMoveablePos.entries())
	return false;

    if (!avoidmob)
    {
	i = rand_choice(myMoveablePos.entries());
	mx = myMoveablePos(i).x;
	my = myMoveablePos(i).y;
	return true;
    }

    // A couple random guesses as this is much faster.
    for (trials = 0; trials < 10; trials++)
    {
	i = rand_choice(myMoveablePos.entries());
	x = myMoveablePos(i).x;
	y = myMoveablePos(i).y;

	if (!getMob(x, y))
	{
	    mx = x;
	    my = y;
	    return true;
	}
    }
    
    // Otherwise loop through..
    for (i = 0; i < myMoveablePos.entries(); i++)
    {
	x = myMoveablePos(i).x;
	y = myMoveablePos(i).y;
	
	if (getMob(x, y))
	    continue;

	nummatch++;
	if (!rand_choice(nummatch))
	{
	    mx = x;
	    my = y;
	}
    }

    if (nummatch)
	return true;
    return false;
}

bool
MAP::findMoveableUnexplored(int &mx, int &my) const
{
    int		nummatch = 0;
    int		x, y;
    
    FORALL_MAP(x, y)
    {
	if (isPassable(x, y) && !isMapped(x, y))
	{
	    nummatch++;
	    if (!rand_choice(nummatch))
	    {
		mx = x;
		my = y;
	    }
	}
    }

    if (nummatch)
	return true;
    return false;
}

bool
MAP::findCloseMoveableUnexplored(int &mx, int &my, int cx, int cy) const
{
    int		nummatch = 0;
    int		x, y;
    int		mindist = 1000, d;
    
    FORALL_MAP(x, y)
    {
	if (isPassable(x, y) && !isMapped(x, y))
	{
	    // Verify we are closer.
	    // If we are equal, increase matches
	    // If we are closer, reset matches.
	    d = dist(x, y, cx, cy);
	    if (d < mindist)
	    {
		nummatch = 1;
		mindist = d;
	    }
	    else if (d == mindist)
	    {
		nummatch++;
	    }
	    else
		continue;
		    
	    if (!rand_choice(nummatch))
	    {
		mx = x;
		my = y;
	    }
	}
    }

    if (nummatch)
	return true;
    return false;
}

bool
MAP::findNextSymbol(char symbol, int &mx, int &my) const
{
    int		nummatch = 0;
    int		x, y;
    TILE_NAMES	tile;
    MOB		*mob;
    ITEM	*item;
    bool	isblind;
    MOB		*avatar;

    avatar = MOB::getAvatar();
    isblind = false;
    if (avatar)
    {
	isblind = avatar->isBlind();
	if (avatar->isSwallowed())
	    isblind = true;
    }
    	

    x = mx;
    y = my;
    while (1)
    {
	// Increment position
	x++;
	if (x >= getWidth())
	{
	    x = 0;
	    y++;
	    if (y >= getHeight())
		y = 0;
	}
	
	// Return failure if back to start.
	if (x == mx && y == my)
	    return false;

	// See if we have the right symbol
	tile = getTile(x, y);
	if (glb_tiledefs[tile].symbol == symbol)
	{
	    mx = x;
	    my = y;
	    return true;
	}

	// See if the monster is correct
	mob = getMob(x, y);
	if (!isblind && mob && mob->defn().symbol == symbol)
	{
	    // Don't see things out of los if hide los is true.
	    if (!mob->defn().hidefromlos || isFOV(x, y))
	    {
		mx = x;
		my = y;
		return true;
	    }
	}    

	// See if the monster is correct
	item = getItem(x, y);
	if (!isblind && item && item->defn().symbol == symbol)
	{
	    mx = x;
	    my = y;
	    return true;
	}    
    }

    // Never reached.
    return false;
}

#define PIECE(x,y) (piece[(y)*piece_w+(x)])

void
MAP::build(MAPSIZE_NAMES size, int numriver, int numfire)
{
    // Build the world...
    int		rooms = 0;

    myWidth = glb_mapsizedefs[size].width;
    myHeight = glb_mapsizedefs[size].height;

    int			piece_w = myWidth / 5;
    int			piece_h = myHeight / 5;
    
    myTiles = MAPDATA(myWidth, myHeight);
    myTileFlags = MAPDATA(myWidth, myHeight);
    myWetMap = MAPDATA(myWidth, myHeight);

    // Clear ourselves to wall.
    clear(TILE_ICEMOUNTAIN);
    clearFlag(MAPFLAG_MAP, false);
    clearFlag(MAPFLAG_FOV, false);

    PIECE		**piece;
    int			x, y, i;
    bool		newpiece;

    piece = new PIECE *[piece_w * piece_h];
    
    memset(piece, 0, sizeof(PIECE *) * piece_w * piece_h);

    PIECEDICT		piecedict;
    int			top, right, bot, left;

    // Initialize center & grow...
    PIECE(piece_w/2, piece_h/2) = piecedict.getRandomPiece(-1, -1, -1, -1);
    
    newpiece = true;
    while (newpiece)
    {
	newpiece = false;
	for (y = 0; y < piece_h; y++)
	{
	    for (x = 0; x < piece_w; x++)
	    {
		if (PIECE(x,y))
		    continue;

		// See if any point continuing.
		if ((x && PIECE(x-1,y) && PIECE(x-1,y)->getCode(CODE_RIGHT) != SOLID_CODE) ||
		    (x < (piece_w-1) && PIECE(x+1,y) && PIECE(x+1,y)->getCode(CODE_LEFT) != SOLID_CODE) ||	
		    (y && PIECE(x,y-1) && PIECE(x,y-1)->getCode(CODE_BOT) != SOLID_CODE) ||	
		    (y < (piece_h-1) && PIECE(x,y+1) && PIECE(x,y+1)->getCode(CODE_TOP) != SOLID_CODE))
		{
		    // Get our codes.
		    if (x)
		    {
			if (PIECE(x-1,y))
			    left = PIECE(x-1,y)->getCode(CODE_RIGHT);
			else
			    left = -1;
		    }
		    else
			left = SOLID_CODE;
		    if (x < piece_w-1)
		    {
			if (PIECE(x+1,y))
			    right = PIECE(x+1,y)->getCode(CODE_LEFT);
			else
			    right = -1;
		    }
		    else
			right = SOLID_CODE;
		    if (y)
		    {
			if (PIECE(x,y-1))
			    top = PIECE(x,y-1)->getCode(CODE_BOT);
			else
			    top = -1;
		    }
		    else
			top = SOLID_CODE;
		    if (y < piece_h-1)
		    {
			if (PIECE(x,y+1))
			    bot = PIECE(x,y+1)->getCode(CODE_TOP);
			else
			    bot = -1;
		    }
		    else
			bot = SOLID_CODE;

		    PIECE(x,y) = piecedict.getRandomPiece(top, right, bot, left);
		    // assert(piece[y][x]);
		    if (PIECE(x,y))
			newpiece = true;
		}
	    }
	}
    }

    // Write out all the pieces.
    for (y = 0; y < piece_h; y++)
    {
	for (x = 0; x < piece_w; x++)
	{
	    if (PIECE(x,y))
	    {
		rooms++;
		PIECE(x,y)->writeToMap(this, piecedict.legend(), x*5, y*5);
	    }
	}
    }

    // Force the outer wall to be wall...
    for (x = 0; x < myWidth; x++)
    {
	setTile(x, 0, TILE_ICEMOUNTAIN);
	setTile(x, myHeight-1, TILE_ICEMOUNTAIN);
    }
    for (y = 0; y < myHeight; y++)
    {
	setTile(0, y, TILE_ICEMOUNTAIN);
	setTile(myWidth-1, y, TILE_ICEMOUNTAIN);
    }

    // Update total room count.
    myTotalRooms = rooms;
    myRooms = rooms;

    // Find the diametrically opposed coordinates
    POSLIST		list;
    POS			a;
    int			nobstacles = numriver + numfire;

    getDiameter(list, nobstacles + 1);

    for (i = 0; i < list.entries(); i++)
    {
	list.set(i, findSetBack(list(i), 2));
    }

    // Avatar gets the last position which is likely in the center.
    a = list.top();
    
    // Start of the flood
    for (i = 0; i < nobstacles; i++)
    {
	// Randomly pick from our river & fire list, removing
	// on selection.
	if (rand_choice(numriver+numfire) < numriver)
	{
	    setTile(list(i), TILE_WATER);
	    numriver--;
	}
	else
	{
	    if (getTile(list(i)) == TILE_GRASS)
		setTile(list(i), TILE_BRUSHFIRE);
	    else
		setTile(list(i), TILE_FORESTFIRE);
	    numfire--;
	}
    }

    // Avatar's location:
    {
	MOB		*avatar = MOB::getAvatar();

	if (!avatar)
	{
	    avatar = MOB::create(MOB_AVATAR);
	    MOB::setAvatar(avatar);
	    addMob(avatar);
	}
	avatar->move(a);
	// Create our children surround us, a safe bet as we 
	// are set back.
	MOB		*child;
	int		 dx, dy;
	POS		 b = list(0);

	// Protectively, the kids are away from the water.
	// We pick the first river at random here.
	dx = SIGN(a.x - b.x);
	if (!dx) dx = 1;
	dy = SIGN(a.y - b.y);
	if (!dy) dy = 1;

	child = MOB::create(MOB_PRIMUS);
	child->move(a.x,a.y+dy);
	addMob(child);
	child = MOB::create(MOB_DEUCE);
	child->move(a.x+dx,a.y);
	addMob(child);
	child = MOB::create(MOB_TREO);
	child->move(a.x+dx,a.y+dy);
	addMob(child);
    }

    // Add items.
    for (i = 0; i < (rooms/2); i++)
    {
	ITEM		*item;

	findMoveable(x, y, true);
	item = ITEM::createRandom(getDepth());

	item->move(0, x, y);
	addItem(item);
    }

    // Store our initial state.
    myOrigState = saveToText();

    clearAllCaches();
}

void
MAP::clear(TILE_NAMES tile)
{
    myTiles.constant(tile);
}

void
MAP::clearFlag(int flag, bool val)
{
    int	    x, y;

    for (y = 0; y < myHeight; y++)
    {
	for (x = 0; x < myWidth; x++)
	{
	    setFlag(x, y, flag, val);
	}
    }
}

void
MAP::buildFOV(int cx, int cy)
{
    // Clear out FOV flag.
    clearFlag(MAPFLAG_FOV, false);

    // Starting with x, y, we spiral outwards so long
    // as we still have valid values.

    int		radius, maxrad;
    bool	hadfov;
    int		i, iminus;

    if (!isValidCoord(cx, cy))
	return;

    // Center is in FOV.
    setFlag(cx, cy, MAPFLAG_FOV|MAPFLAG_MAP, true);

    maxrad = MAX(myWidth, myHeight);
    for (radius = 1; radius < maxrad; radius++)
    {
	hadfov = false;
	for (i = -radius; i <= radius; i++)
	{
	    // This is i, but one step closer to the center.
	    iminus = i - SIGN(i);

	    if (isFOV(cx + i, cy + (radius-1)) ||
		isFOV(cx + iminus, cy + (radius-1)))
	    {
		if (hasLOSLoose(cx, cy, cx+i, cy+radius))
		{
		    setFlag(cx+i, cy+radius, MAPFLAG_FOV|MAPFLAG_MAP, true);
		    hadfov = true;
		}
	    }
	    if (isFOV(cx + i, cy - (radius-1)) ||
		isFOV(cx + iminus, cy - (radius-1)))
	    {
		if (hasLOSLoose(cx, cy, cx+i, cy-radius))
		{
		    setFlag(cx+i, cy-radius, MAPFLAG_FOV|MAPFLAG_MAP, true);
		    hadfov = true;
		}
	    }
	    if (isFOV(cx + (radius-1), cy + i) ||
		isFOV(cx + (radius-1), cy + iminus))
	    {
		if (hasLOSLoose(cx, cy, cx+radius, cy+i))
		{
		    setFlag(cx+radius, cy+i, MAPFLAG_FOV|MAPFLAG_MAP, true);
		    hadfov = true;
		}
	    }
	    if (isFOV(cx - (radius-1), cy + i) ||
		isFOV(cx - (radius-1), cy + iminus))
	    {
		if (hasLOSLoose(cx, cy, cx-radius, cy+i))
		{
		    setFlag(cx-radius, cy+i, MAPFLAG_FOV|MAPFLAG_MAP, true);
		    hadfov = true;
		}
	    }
	}

	// Early exit if everything in last circle was
	// out of fov.
	if (!hadfov)
	    break;
    }
}

bool
MAP::hasLOSLoose(int sx, int sy, int ex, int ey) const
{
    int		dx, dy;

    for (dy = -1; dy <= 1; dy++)
    {
	for (dx = -1; dx <= 1; dx++)
	{
	    if ((!dx && !dy) || isTransparent(sx+dx, sy+dy))
	    {
		if (hasLOS(sx+dx, sy+dy, ex, ey))
		    return true;
	    }
	}
    }

    return false;
}

bool
MAP::hasLOS(int sx, int sy, int ex, int ey) const
{
    int		cx, cy;
    int		acx, acy;
    int		x, y;

    cx = ex - sx;
    cy = ey - sy;
    acx = ABS(cx);
    acy = ABS(cy);

    // Invalid coords are never in LOS.
    if (!isValidCoord(sx, sy))
	return false;
    if (!isValidCoord(ex, ey))
	return false;
    
    // Check for trivial LOS.
    if (acx <= 1 && acy <= 1)
	return true;

    // We do a bresenham algorithm.
    // Note that we do *NOT* check the start or
    // the end coords!  Something is in LOS despite the value
    // of these coords.
    if (acx > acy)
    {
	// X major stepping.
	int		dx, dy, error;

	dx = SIGN(cx);
	dy = SIGN(cy);

	error = 0;
	
	// Take initial free step.
	error = acy;
	y = sy;
	for (x = sx + dx; x != ex; x += dx)
	{
	    if (error >= acx)
	    {
		error -= acx;
		y += dy;
	    }

	    // The current ray is between (x, y) and (x, y+dy).
	    // Thus, if error is not zero, we consider it a pass
	    // if *either* is transparent.
	    // Whoops. This gives tunnel vision!  You can tunnel
	    // through one thickness walls.
	    if (!isTransparent(x, y) ) // &&
//		(error == 0 || !isTransparent(x, y+dy)))    
		return false;

	    error += acy;
	}
    }
    else
    {
	// Y major stepping.
	int		dx, dy, error;

	dx = SIGN(cx);
	dy = SIGN(cy);

	error = 0;
	
	// Take initial free step.
	error = acx;
	x = sx;
	for (y = sy + dy; y != ey; y += dy)
	{
	    if (error >= acy)
	    {
		error -= acy;
		x += dx;
	    }

	    // The current ray is between (x, y) and (x+dx, y).
	    // Thus, if error is not zero, we consider it a pass
	    // if *either* is transparent.
	    if (!isTransparent(x, y) ) // &&
//		(error == 0 || !isTransparent(x + dx, y)))
		return false;

	    error += acx;
	}
    }

    // Nothing got in our way, so we're done.
    return true;
}

MAPDATA
MAP::buildDistMap(POS start) const
{
    int			dx, dy, dist;
    POS			p, np;
    POSLIST		stack;
    MAPDATA		result(getWidth(), getHeight());

    // First, clear everything out.
    result.constant(-1);

    // No distance to root..
    result.data(start.x, start.y) = 0;

    stack.append(start);
    while (stack.entries())
    {
	p = stack.removeFirst();

	dist = result(p.x, p.y);
	dist++;

	FORALL_8DIR(dx, dy)
	{
	    np.x = p.x+dx;
	    np.y = p.y+dy;

	    if (!isValidCoord(np))
		continue;

	    // Ignore impassable
	    if (!isPassable(np))
		continue;

	    // Ignore already set
	    if (result(np.x, np.y) >= 0)
		continue;

	    // Calculate distance
	    result.data(np.x, np.y) = dist;

	    // And add to stack.
	    stack.append(np);
	}
    }

    return result;
}


void
MAP::getDiameter(POSLIST &list, int npt, bool quiet) const
{
    // Finding the actual diameter is too much for a 7drl.
    // Instead, I shall use a simple heuristic.  Picking a random
    // point, I find the deepest point from there.  That becomes a.
    // I then find the deepest point from a to get b.  There is a non
    // zero chance this algorithma actually produces the diameter
    // but I don't want to spend the time trying to prove/disprove it.
    //
    // I am very certain this is not optimal for npt > 2
    POS			p;

    list.clear();

    while (list.entries() < npt)
    {
	p = findDeepest(list);
	list.append(p);

	if (!quiet)
	{
	    // This is a slow operation...
	    msg_report("#");
	    msg_update();
	    gfx_update();
	}
    }
}

POS
MAP::findDeepest(POS a) const
{
    POS		p, deepest;
    int		depth, maxdepth = -1, numcollide; 

    numcollide = 0;
    FORALL_MAPPOS(p)
    {
	// Only consider connected coords
	if (!isConnected(a, p))
	    continue;

	depth = getDist(a, p);
	if (depth == maxdepth)
	{
	    numcollide++;
	    // Use this point at random.
	    if (!rand_choice(numcollide))
		maxdepth--;
	}
	else if (depth > maxdepth)
	    numcollide = 1;
	
	if (depth > maxdepth)
	{
	    deepest = p;
	    maxdepth = depth;
	}
    }

    return deepest;
}

POS
MAP::findDeepest(const POSLIST &list) const
{
    POS		p, deepest;
    int		i, mindist, maxdist = -1, dist;
    int		numcollide = 0;
    
    // By default just find any leaf node.
    if (!list.entries())
    {
	findMoveable(p.x, p.y);
	return findDeepest(p);
    }

    // List of all of our distance maps
    // Caching them here avoids us having to look up every time,
    // important when we have 100 maps since our LRU cache will be 
    // exactly wrong.
    PTRLIST<MAPDATA>	maplist;

    for (i = 0; i < list.entries(); i++)
    {
	maplist.append(getDistMap(list(i)));
    }

    FORALL_MAPPOS(p)
    {
	mindist = 1000;
	for (i = 0; i < maplist.entries(); i++)
	{
	    dist = maplist(i)(p.x, p.y);
	    if (dist < mindist)
		mindist = dist;
	}
	// Ignore unreachable.
	if (mindist == -1)
	    continue;

	if (mindist == maxdist)
	{
	    numcollide++;
	    if (!rand_choice(numcollide))
		maxdist--;
	}
	else if (mindist > maxdist)
	    numcollide = 1;
	
	// See if this is a new maximum.
	if (mindist > maxdist)
	{
	    maxdist = mindist;
	    deepest = p;
	}
    }

    return deepest;
}

POS
MAP::findSetBack(POS a, int dist) const
{
    POS		p, closest;
    int		depth, mindepth = 1000, numcollide; 

    closest = a;
    numcollide = 0;
    FORALL_MAPPOS(p)
    {
	// Only consider connected coords
	if (!isConnected(a, p))
	    continue;

	// Only consider those that are valid
	if (wallDist(p) < dist)
	    continue;

	depth = getDist(a, p);
	if (depth == mindepth)
	{
	    numcollide++;
	    // Use this point at random.
	    if (!rand_choice(numcollide))
		mindepth++;
	}
	else if (depth < mindepth)
	    numcollide = 1;
	
	if (depth < mindepth)
	{
	    closest = p;
	    mindepth = depth;
	}
    }

    return closest;
}

bool
MAP::isConnected(POS p1, POS p2, int depth) const
{
    if (!isValidCoord(p1))
	return false;
    if (!isValidCoord(p2))
	return false;
    if (myConnectivityDepth != depth)
    {
	myConnectivity = buildConnectivity(depth);
	myConnectivityDepth = depth;
    }
    int		c1, c2;

    c1 = findClass(myConnectivity, p1.x, p1.y);
    c2 = findClass(myConnectivity, p2.x, p2.y);
    if (c1 == -1)
	return false;
    return c1 == c2;
}



void
MAP::mapConnected(POS p)
{
    POS		p2;
    int		dx, dy;

    FORALL_MAPPOS(p2)
    {
	if (isConnected(p, p2))
	    setFlag(p2, MAPFLAG_MAP, true);
	else
	{
	    // Check if any adjacent dudes mapped
	    FORALL_8DIR(dx, dy)
	    {
		if (isConnected(p, POS(p2.x+dx, p2.y+dy)))
		{
		    setFlag(p2, MAPFLAG_MAP, true);
		    break;
		}
	    }
	}
    }
}

void
MAP::clearAllCaches()
{
    myConnectivityDepth = -1;
    myWaterDepthValid = false;
    myWallDistValid = false;
    myDistMapPos.clear();
    myDistMaps.clear();
    myMoveablePos.clear();
}

int
MAP::waterDepth(int x, int y) const
{
    if (!myWaterDepthValid)
    {
	myWaterDepth = buildWaterDepth();
	myWaterDepthValid = true;
    }

    if (myWaterDepth(x, y) < 0)
    {
	// Unreachable, so infinite depth
	return 1000;
    }
    return myWaterDepth(x, y);
}

int
MAP::wallDist(POS p) const
{
    if (!myWallDistValid)
    {
	myWallDist = buildWallDist();
	myWallDistValid = true;
    }

    return myWallDist(p.x, p.y);
}


MAPDATA
MAP::buildWaterDepth() const
{
    MAPDATA 	result(getWidth(), getHeight());
    POSLIST	front;
    int		x, y, dx, dy, nx, ny;
    
    result.constant(-1);

    // Breadth first search marking tiles according to how far they
    // are from water.
    // We mark distances with 8 way connectivity as that is how
    // water flows.  However, we want to only mark connected
    // if water can actually flow in that direction!
    // This does mean standing behind a dam is safe, well, so long as the
    // dam is water tight.
    // Note, however, that standing *on* a dam should be marked as unsafe
    // if any of the four directions is water!

    // Initialize the wave front with all water tiles.
    FORALL_MAP(x, y)
    {
	if (glb_tiledefs[getTile(x, y)].isdangerous)
	{
	    front.append(POS(x, y));
	    // Trivially zero distance.
	    result.data(x,y) = 0;
	}
	else if (getItem(x, y))
	{
	    // Check to see if boredred by water...
	    FORALL_4DIR(dx, dy)
	    {
		if (glb_tiledefs[getTile(x+dx,y+dy)].isdangerous)
		{
		    // This is a dam.  We want dams to always be unsafe, but
		    // note we don't want to propagate from dams.
		    // A nice side effect is the kids won't displace
		    // boulders so long as you haven't acted unsafe.
		    result.data(x,y) = 1;
		}
	    }
	}
    }

    // Now run through our wave front breadth first.
    while (front.entries())
    {
	POS		p = front.removeFirst();
	x = p.x;
	y = p.y;

	FORALL_8DIR(dx, dy)
	{
	    nx = x + dx;
	    ny = y + dy;
	    
	    if (!isValidCoord(nx, ny))
		continue;
	    
	    // Breadth first means that if we are not unassigned,
	    // we are already closer
	    if (result(nx, ny) != -1)
		continue;

	    if (!isWettable(nx, ny))
		continue;
	    // Likewise, verify that there exists a traversal path.
	    if (dx && dy)
	    {
		if (!isWettable(nx,y) && !isWettable(x,ny))
		    continue;
	    }

	    // We have a new path, append to our search list.
	    front.append(POS(nx, ny));
	    result.data(nx, ny) = result(x, y) + 1;
	}
    }

    return result;
}

MAPDATA
MAP::buildWallDist() const
{
    MAPDATA 	result(getWidth(), getHeight());
    POSLIST	front;
    int		x, y, dx, dy, nx, ny;
    
    result.constant(-1);

    // Straightforward distance from a wall.

    // Initialize the wave front with all water tiles.
    FORALL_MAP(x, y)
    {
	if (getTile(x, y) == TILE_MOUNTAIN ||
	    getTile(x, y) == TILE_ICEMOUNTAIN)
	{
	    front.append(POS(x, y));
	    // Trivially zero distance.
	    result.data(x,y) = 0;
	}
    }

    // Now run through our wave front breadth first.
    while (front.entries())
    {
	POS		p = front.removeFirst();
	x = p.x;
	y = p.y;

	FORALL_8DIR(dx, dy)
	{
	    nx = x + dx;
	    ny = y + dy;
	    
	    if (!isValidCoord(nx, ny))
		continue;
	    
	    // Breadth first means that if we are not unassigned,
	    // we are already closer
	    if (result(nx, ny) != -1)
		continue;

	    // We have a new path, append to our search list.
	    front.append(POS(nx, ny));
	    result.data(nx, ny) = result(x, y) + 1;
	}
    }

    return result;
}

int
MAP::getDist(POS p1, POS p2) const
{
    if (!isValidCoord(p1))
	return -1;
    if (!isValidCoord(p2))
	return -1;

    // Check to see if we already have a distance map...
    int			i;
    for (i = 0; i < myDistMapPos.entries(); i++)
    {
	if (p2 == myDistMapPos(i))
	    return myDistMaps(i)(p1.x, p1.y);

	// p1 is considered the serious cache so we swap it
	// in.
	if (p1 == myDistMapPos(i))
	{
	    myDistMaps.swapEntries(0, i);
	    myDistMapPos.swapEntries(0, i);
	    return myDistMaps(0)(p2.x, p2.y);
	}
    }
    
    MAPDATA	dist;

    dist = buildDistMap(p1);

    //  Hmm...  Probably should prepend, but then again
    // that swap should happen on the next access anyways.
    myDistMapPos.append(p1);
    myDistMaps.append(dist);
    return dist(p2.x, p2.y);
}

MAPDATA
MAP::getDistMap(POS p) const
{
    assert(isValidCoord(p));

    // Check to see if we already have a distance map...
    int			i;
    for (i = 0; i < myDistMapPos.entries(); i++)
    {
	// We don't run LRU here as the caller is implicitly caching.
	if (p == myDistMapPos(i))
	    return myDistMaps(i);
    }

    // Non exists, so create. 
    MAPDATA	dist;

    dist = buildDistMap(p);

    //  Hmm...  Probably should prepend, but then again
    // that swap should happen on the next access anyways.
    myDistMapPos.append(p);
    myDistMaps.append(dist);

    return dist;
}

bool
MAP::isValidCoord(int x, int y) const
{
    if (x < 0 || x >= myWidth)
	return false;

    if (y < 0 || y >= myHeight)
	return false;

    return true;
}

TILE_NAMES
MAP::getTile(int x, int y) const
{
    if (!isValidCoord(x, y))
	return TILE_INVALID;
    
    return (TILE_NAMES) myTiles(x, y);
}

void
MAP::setTile(int x, int y, TILE_NAMES tile)
{
    assert(isValidCoord(x, y));
    if (!isValidCoord(x, y))
	return;

    // This likely invalidates our distance map assumptions
    clearAllCaches();
    myTiles.data(x, y) = tile;
}

int
MAP::getWetness(int x, int y) const
{
    if (!isValidCoord(x, y))
	return 0;

    return myWetMap(x, y);
}

void
MAP::setWetness(int x, int y, int value)
{
    if (!isValidCoord(x, y))
	return;

    // Clamp into valid range.
    // We allow very large values here because:
    // a) we are an int anyways
    // b) users may try playing with the transition points.
    if (value > 1000)
	value = 1000;
    if (value < -1000)
	value = -1000;
    myWetMap.data(x, y) = value;
}

void
MAP::addWetness(int x, int y, int amount)
{
    if (!isValidCoord(x, y))
	return;

    int		wetness;
    wetness = getWetness(x, y);
    wetness += amount;
    setWetness(x, y, wetness);
}

bool
MAP::isPassable(int x, int y) const
{
    TILE_NAMES	tile;

    tile = getTile(x, y);

    return glb_tiledefs[tile].ispassable;
}

bool
MAP::isWettable(int x, int y) const
{
    if (!isValidCoord(x, y))
	return false;

    TILE_NAMES		tile = getTile(x, y);

    if (tile != TILE_WATER &&
	glb_tiledefs[tile].wettile == TILE_INVALID)
	return false;
    if (getItem(x, y))
	return false;

    return true;
}

bool
MAP::isBurnable(int x, int y) const
{
    if (!isValidCoord(x, y))
	return false;

    TILE_NAMES		tile = getTile(x, y);

    if (glb_tiledefs[tile].drytile == TILE_INVALID)
	return false;
    if (getItem(x, y))
	return false;

    return true;
}

bool
MAP::isTransparent(int x, int y) const
{
    TILE_NAMES	tile;

    tile = getTile(x, y);

    return glb_tiledefs[tile].istransparent;
}

bool
MAP::isMapped(int x, int y) const
{
    return getFlag(x, y, MAPFLAG_MAP);
}

bool
MAP::isFOV(int x, int y) const
{
    return getFlag(x, y, MAPFLAG_FOV);
}

ITEM *
MAP::getItem(int x, int y) const
{
    ITEM		*item;

    if (!hasItem(x, y))
	return 0;

    for (item = myItems; item; item = item->getNext())
    {
	if (item->getX() == x && item->getY() == y)
	    return item;
    }
    setFlag(x, y, MAPFLAG_ITEM, false);
    return 0;
}

bool
MAP::hasItem(int x, int y) const
{
    if (!isValidCoord(x, y))
	return false;

    return getFlag(x, y, MAPFLAG_ITEM);
}

void
MAP::getItemList(ITEMLIST &list, int x, int y) const
{
    list.clear();

    if (!hasItem(x, y))
	return;

    ITEM		*item;

    for (item = myItems; item; item = item->getNext())
    {
	if (item->getX() == x && item->getY() == y)
	    list.append(item);
    }
}

MOB *
MAP::getMob(int x, int y) const
{
    MOB		*mob;

    if (!hasMob(x, y))
	return 0;

    for (mob = myMobs; mob; mob = mob->getNext())
    {
	// Swallowed mobs not actually on the map.
	if (mob->isSwallowed())
	    continue;
	if (mob->getX() == x && mob->getY() == y)
	    return mob;
    }

    // Obviously map is out of date
    setFlag(x, y, MAPFLAG_MOB, false);
    return 0;
}

bool
MAP::hasMob(int x, int y) const
{
    if (!isValidCoord(x, y))
	return false;

    return getFlag(x, y, MAPFLAG_MOB);
}

int
MAP::getNumMobs() const
{
    int		num = 0;
    MOB		*mob;

    for (mob = myMobs; mob; mob = mob->getNext())
    {
	num++;
    }
    return num;
}
int
MAP::getAllMobs(PTRLIST<MOB *> &moblist, int x, int y) const
{
    MOB		*mob;

    for (mob = myMobs; mob; mob = mob->getNext())
    {
	if (mob->getX() == x && mob->getY() == y)
	{
	    moblist.append(mob);
	}
    }
    return moblist.entries();
}

MOB *
MAP::findMob(MOB_NAMES def) const
{
    MOB		*mob;

    for (mob = myMobs; mob; mob = mob->getNext())
    {
	if (mob->getDefinition() == def)
	    return mob;
    }
    return 0;
}

MOB *
MAP::traceBullet(int x, int y, int range, int dx, int dy) const
{
    MOB		*hit = 0;

    x += dx;
    y += dy;
    while (isValidCoord(x, y))
    {
	hit = getMob(x, y);
	if (hit)
	    break;

	if (!isTransparent(x, y))
	    break;
	
	// Move one step
	x += dx;
	y += dy;
	range--;
	if (range <= 0)
	    break;
    }

    return hit;
}

void
MAP::displayBullet(int x, int y, int range, int dx, int dy, u8 sym, ATTR_NAMES attr) const
{
    MOB		*hit = 0;
    int		 sx, sy;

    x += dx;
    y += dy;
    while (isValidCoord(x, y))
    {
	hit = getMob(x, y);

	if (!isTransparent(x, y))
	    break;

	// No delay for out of sight changes.
	if (isFOV(x, y))
	{
	    re_redraw();

	    // Plot our char.
	    sx = x + myLastX - myLastMX;
	    sy = y + myLastY - myLastMY;
	    // Ensure on screen...
	    if (sx >= myLastX && sx < myLastX + myLastW &&
		sy >= myLastY && sy < myLastY + myLastH)
	    {
		gfx_printchar(sx, sy, sym, attr);
	    }
	    
	    gfx_update();
	    gfx_delay(5);
	}
	
	// Move one step
	x += dx;
	y += dy;
	range--;
	if (range <= 0)
	    break;
	if (hit)
	    break;
    }

    // clear up display.
    re_redraw();
    gfx_update();
}

void
MAP::fireball(MOB *src, int tx, int ty, int rad, DPDF damage, 
	      u8 sym, ATTR_NAMES attr) 
{
    MOB		*hit = 0;
    int		 x, y, sx, sy;
    bool	 anyvis = false;

    re_redraw();
    for (y = ty-rad+1; y <= ty+rad-1; y++)
    {
	for (x = tx-rad+1; x <= tx+rad-1; x++)
	{
	    // No delay for out of sight changes.
	    if (!isValidCoord(x, y))
		continue;
	    if (!isFOV(x, y))
		continue;
	    
	    anyvis = true;

	    // Plot our char.
	    sx = x + myLastX - myLastMX;
	    sy = y + myLastY - myLastMY;
	    // Ensure on screen...
	    if (sx >= myLastX && sx < myLastX + myLastW &&
		sy >= myLastY && sy < myLastY + myLastH)
	    {
		gfx_printchar(sx, sy, sym, attr);
	    }
	}
    }
	    
    if (anyvis)
    {
	gfx_update();
    }
    for (y = ty-rad+1; y <= ty+rad-1; y++)
    {
	for (x = tx-rad+1; x <= tx+rad-1; x++)
	{
	    if (!isValidCoord(x, y))
		continue;

	    hit = getMob(x, y);
	    
	    // Ignore self hits.
	    if (hit == src)
		continue;

	    if (hit)
		hit->applyDamage(src, damage.evaluate());
	}
    }

    // Delay after damage effect
    if (anyvis)
    {
	// Only do extra delay if explosion
	if (rad > 1)
	    gfx_delay(8);
	re_redraw();
	gfx_update();
    }
}

void
MAP::describeSquare(int x, int y, bool isblind) const
{
    BUF		 buf;
    buf = getTerrainName(x, y);
    if (!isblind)
    {
	MOB		*mob;
	ITEMLIST	 list;
	int		 depth;

	mob = getMob(x, y);
	msg_format("%S <see> %O.", MOB::getAvatar(), buf);
	
	depth = waterDepth(x, y);
	if (depth < 500)
	    buf.sprintf("Danger is %d meter%s away.", depth, (depth == 1 ? "" : "s"));
	else
	    buf.reference("This place is safe.");
	msg_report(buf);

	if (mob)
	    msg_format("%S <see> %O.", MOB::getAvatar(), mob);
	getItemList(list, x, y);
	if (list.entries())
	{
	    msg_format("%S <see> ", MOB::getAvatar());
	    for (int i = 0; i < list.entries(); i++)
	    {
		if (i)
		    msg_report(", ");
		msg_report(list(i)->getName(true));
	    }
	    msg_report(".");
	}
    }
    else if (isMapped(x, y))
    {
	msg_format("%S <recall> %O.", MOB::getAvatar(), buf);
    }
    else
    {
	msg_format("%S <know> nothing of that spot.", MOB::getAvatar());
    }
}

BUF
MAP::getTerrainName(int x, int y) const
{
    TILE_NAMES		tile;
    BUF			result;
    const char 		*soggy;
    float		ratio;
    int			wetvalue;
    const int		numwetness = 5;
    const char 		*soggynames[numwetness] =
    {
	"damp ",
	"wet ",
	"soggy ",
	"sodden ",
	"waterlogged "
    };
    const int		numdryness = 5;
    const char 		*drynames[numdryness] =
    {
	"dry ",
	"baked ",
	"parched ",
	"bone-dry ",
	"tinder-like "
    };

    tile = getTile(x, y);

    if (getWetness(x, y) > 0)
    {
	// Determine how soggy this tile is.
	if (glb_tiledefs[tile].wettile == TILE_INVALID)
	    soggy = "";		// Immutable
	else
	{
	    ratio = ((float)getWetness(x, y)) / glb_tiledefs[tile].wettransition;
	    wetvalue = (int) (ratio * numwetness);
	    if (wetvalue >= numwetness)
		wetvalue = numwetness-1;
	    soggy = soggynames[wetvalue];
	}
    }
    else if (getWetness(x, y) < 0)
    {
	// Determine how soggy this tile is.
	if (glb_tiledefs[tile].drytile == TILE_INVALID)
	    soggy = "";		// Immutable
	else
	{
	    ratio = ((float)getWetness(x, y)) / glb_tiledefs[tile].drytransition;
	    wetvalue = (int) (ratio * numdryness);
	    if (wetvalue >= numdryness)
		wetvalue = numdryness-1;
	    soggy = drynames[wetvalue];
	}
    }
    else
	soggy = "";	// Plain.
    

    result.sprintf("%s%s", soggy, glb_tiledefs[tile].legend);

    return result;
}

void
MAP::addItem(ITEM *item)
{
    assert(item->getNext() == 0);

    item->setNext(myItems);
    myItems = item;

    setFlag(item->getPos(), MAPFLAG_ITEM, true);
}

void
MAP::removeItem(ITEM *item)
{
    ITEM	*c, *l;

    assert(item);
    if (!item)
	return;

    l = 0;
    for (c = myItems; c; c = c->getNext())
    {
	if (c == item)
	{
	    if (l)
		l->setNext(item->getNext());
	    else
		myItems = item->getNext();
	    item->setNext(0);
	    return;
	}

	l = c;
    }	    

    assert(0);
}

void
MAP::addMob(MOB *mob)
{
    assert(mob->getNext() == 0);

    mob->setNext(myMobs);
    myMobs = mob;

    mob->setMap(this);

    setFlag(mob->getPos(), MAPFLAG_MOB, true);
    myEverHad[mob->getDefinition()] = true;
}

void
MAP::removeMob(MOB *mob)
{
    MOB		*cur, *last;

    // Protect our safe next flag.
    if (mob && mob == myMobSafeNext)
    {
	myMobSafeNext = mob->getNext();
    }	    

    last = 0;
    for (cur = myMobs; cur; cur = cur->getNext())
    {
	if (cur == mob)
	{
	    if (last)
		last->setNext(cur->getNext());
	    else
		myMobs = cur->getNext();
	    cur->setNext(0);

	    cur->setMap(0);

	    return;
	}

	last = cur;
    }

    // Failure to remove a mob!
    assert(0);
}

bool
MAP::getFlag(int x, int y, int flag) const
{
    if (!isValidCoord(x, y))
	return false;

    if (myTileFlags(x, y) & flag)
	return true;
    return false;
}

void
MAP::setFlag(int x, int y, int flag, bool val) const
{
    if (!isValidCoord(x, y))
	return;

    if (val)
	myTileFlags.data(x, y) |= flag;
    else
	myTileFlags.data(x, y) &= ~flag;
}

void
MAP::redraw(int sx, int sy, int w, int h,
	    int cx, int cy) const
{
    // Adjust cx & cy so they refer to top left.
    cx = cx - w / 2;
    cy = cy - h / 2;

    myLastX = sx;
    myLastY = sy;
    myLastMX = cx;
    myLastMY = cy;
    myLastW = w;
    myLastH = h;

    re_redraw();
}

void
MAP::re_redraw() const
{
    int		sx, sy, w, h;
    int		x, y, cx, cy;
    TILE_NAMES	tile;
    u8		symbol;
    ATTR_NAMES	attr;
    bool	isblind;
    bool	isswallowed;
    MAPDATA	doubleitem;

    sx = myLastX;
    sy = myLastY;
    w = myLastW;
    h = myLastH;
    cx = myLastMX;
    cy = myLastMY;

    doubleitem = MAPDATA(w, h);
    doubleitem.constant(0);

    isblind = false;
    if (MOB::getAvatar())
	isblind = MOB::getAvatar()->isBlind();

    isswallowed = false;
    if (MOB::getAvatar())
	isswallowed = MOB::getAvatar()->isSwallowed();

    // Plot all the points.
    for (y = sy; y < sy + h; y++)
    {
	for (x = sx; x < sx + w; x++)
	{
	    tile = getTile(x - sx + cx, y - sy + cy);

	    // Plot relevant character.
	    if (x >= 0 && x <= SCR_WIDTH && y >= 0 && y <= SCR_WIDTH)
	    {
		symbol = glb_tiledefs[tile].symbol;
		attr = (ATTR_NAMES) glb_tiledefs[tile].attr;

#if 0
		// Override if out of FOV - shows up dim then.
		if (!isFOV(x - sx + cx, y - sy + cy))
		{
		    // We never draw the stairs dark as the user
		    // should be able to find them easily.
		    if (symbol != '>' && symbol != '<')
			attr = ATTR_OUTOFFOV;
		}
		if (isblind || isswallowed)
		    attr = ATTR_OUTOFFOV;

#endif
		// If it is not mapped, always ' '.
		if (!isMapped(x - sx + cx, y - sy + cy))
		    symbol = ' ';
		
		gfx_printchar(x, y, symbol, attr);
	    }
	}
    }

    // Print out all of our items.
    ITEM	*item;

    for (item = myItems; item; item = item->getNext())
    {
	x = item->getX() + sx - cx;
	y = item->getY() + sy - cy;

	// Verify it is in FOV.
#if 0
	if (!isFOV(item->getX(), item->getY()))
	    continue;
#endif

	if (isblind || isswallowed)
	    continue;

	// Verify it is on screen.
	if (x >= sx && y >= sy && x < sx + w && y < sy + h)
	{
	    doubleitem.data(x-sx, y-sy) = 1 + doubleitem(x-sx, y-sy);
	    item->getLook(symbol, attr);
	    if (doubleitem(x-sx, y-sy) > 1)
	    {
		// Stacked objects show up white so you can tell
		// where extra boulders are.
		attr = ATTR_WHITE;
	    }
	    gfx_printchar(x, y, symbol, attr);
	}
    }

    // Print out all of our mobs.
    MOB		*mob;
    
    for (mob = myMobs; mob; mob = mob->getNext())
    {
	x = mob->getX() + sx - cx;
	y = mob->getY() + sy - cy;

	// Verify it is in FOV.
	if (mob->defn().hidefromlos && !isFOV(mob->getX(), mob->getY()))
	{
	    continue;
	}

	if ((isblind || isswallowed) && !mob->isAvatar())
	    continue;

	// Mobs that are swallowed don't show up.
	if (mob->isSwallowed() && !mob->isAvatar())
	    continue;

	// Verify it is on screen.
	if (x >= sx && y >= sy && x < sx + w && y < sy + h)
	{
	    mob->getLook(symbol, attr);
	    gfx_printchar(x, y, symbol, attr);
	}
    }

    if (isswallowed)
    {
	// Print out the intestine walls.
	x = MOB::getAvatar()->getX() + sx - cx;
	y = MOB::getAvatar()->getY() + sy - cy;
	gfx_printchar(x-1, y-1, '/', ATTR_BROWN);
	gfx_printchar(x,   y-1, '-', ATTR_BROWN);
	gfx_printchar(x+1, y-1, '\\', ATTR_BROWN);
	gfx_printchar(x-1, y,   '|', ATTR_BROWN);
	gfx_printchar(x+1, y,   '|', ATTR_BROWN);
	gfx_printchar(x-1, y+1, '\\', ATTR_BROWN);
	gfx_printchar(x,   y+1, '-', ATTR_BROWN);
	gfx_printchar(x+1, y+1, '/', ATTR_BROWN);
    }
}

void
MAP::fadeToBlack() const
{
    int		r, x, y, d;
    u8		sym;
    ATTR_NAMES	attr;

    d = MAX(myLastW, myLastH) / 2;
    for (r = d+10; r > 0;  r--)
    {
	for (y = 0; y < myLastH; y++)
	{
	    for (x = 0; x < myLastW; x++)
	    {
		d = MAX(ABS(x - myLastW/2), ABS(y - myLastH/2));

		if (d == 0)
		    continue;
		if (d >= r)
		{
		    // pure black.
		    gfx_printchar(x+myLastX, y+myLastY, ' ', ATTR_NORMAL);
		}
		else if (d >= r-10)
		{
		    // Out of FOV.
		    gfx_getchar(x+myLastX, y+myLastY, sym, attr);
		    gfx_printchar(x+myLastX, y+myLastY, sym, ATTR_OUTOFFOV);
		}
	    }
	}
	// Refresh the screen
	gfx_update();

	// Abort on key press...
	if (gfx_isKeyWaiting())
	    break;

	// And delay a bit
	gfx_delay(3);
    }
}

void
MAP::doMoveNPC()
{
    MOB		*cur;

    for (cur = myMobs; cur; cur = myMobSafeNext)
    {
	myMobSafeNext = cur->getNext();

	// I forgot this line the first time.  That was embarrasing :>
	if (cur->isAvatar())
	    continue;
	
	if (!cur->aiForcedAction())
	{
	    buildFOV(cur->getX(), cur->getY());
	    cur->aiDoAI();
	}
    }
}

bool
MAP::flowLiquids(bool forceflow)
{
    // Propagate all tile's wetness to their neighbours
    int		dx, dy;
    int		x, y, j;
    int		wetpropagate;
    TILE_NAMES	tile, wettile;
    bool	steadystate = true;
    MAPDATA	oldwetmap;
    MAPDATA	live;
    ITEM	*item, *nextitem;

    // Water seepage is on slow phases, this ensures everyone
    // can react to the spread.
    if (!forceflow && spd_getphase() != PHASE_NORMAL)
	return false;

    // We try and determine if the setup could change..
    // To determine this, we check our wetmap before & after
    // If any wet values increased on tiles that have a transition,
    // we are not at steady state.  (This allows fires to reach steady
    // state)
    oldwetmap = myWetMap;

    live = buildWaterConnectivity();

    FORALL_MAP(x, y)
    {
	tile = getTile(x, y);
	wetpropagate = glb_tiledefs[tile].wetpropagate;

	// Positive wetpropagate only works for live tiles
	if (wetpropagate > 0)
	{
	    if (!live(x, y))
		continue;
	}

	// This esures fores fires burn out.
	if (glb_tiledefs[tile].selfpropagate)
	{
	    addWetness(x, y, glb_tiledefs[tile].selfpropagate);
	}
	
	if (wetpropagate)
	{
	    for (j = 0; j < 8; j++)
	    {
		rand_angletodir(j, dx, dy);
		if (isValidCoord(x+dx, y+dy))
		{
		    // Check to see if there is a blocker item.
		    // If there is an item at the destination,
		    // we don't wet it.
		    // If there is an item at *both* our ortho
		    // directions, we don't wet.
		    // We check for wettability so you can use
		    // mountains as barriers.
		    if (!isWettable(x+dx, y+dy))
			continue;
		    if (dx && dy)
		    {
			if (!isWettable(x+dx,y) && !isWettable(x,y+dy))
			    continue;
		    }
		}
		addWetness(x+dx, y+dy, wetpropagate);
	    }
	}
    }

    // Check for any items surrounded by water.  We want these
    // to wash away.  The rule is that you need to be supported on a 
    // four way direction by something that *could* dissolve.
    // We don't do anything equivalent for fire as it is a valid tactic
    // to wait for fire to go past you.
    // We do this prior to the transition so a boulder doesn't disappear
    // immediately after being surrounded but lives for a bit.
    // To prevent pairs of boulders from supporting themselves, we only
    // count as land tiles at risk.  Sadly this means a "strong" dam
    // with solid sides has the corners wash away.
    for (item = getItemHead(); item; item = nextitem)
    {
	bool		haswater = false;
	bool		hasland = false;

	nextitem = item->getNext();
	
	x = item->getX();
	y = item->getY();

	FORALL_4DIR(dx, dy)
	{
	    tile = getTile(x+dx, y+dy);
	    if (tile == TILE_WATER)
		haswater = true;
	    else if (isWettable(x+dx,y+dy))
		hasland = true;
	}

	// Special case of the rock or boulder being on a water square
	// This occurs if someone dies on water while carrying it.
	if (getTile(x,y) == TILE_WATER)
	{
	    haswater = true;
	    hasland = false;
	}
	
	if (haswater && !hasland)
	{
	    msg_format("%S <wash> away.", item);
	    // Counts as a transition
	    steadystate = false;
	    removeItem(item);
	    delete item;
	}
    }

    // Having propagated all of the wetness we can, we now transition any
    // tiles so marked.
    // We include non-live tiles as they shouldn't change anyways.
    FORALL_MAP(x, y)
    {
	tile = getTile(x, y);

	wettile = (TILE_NAMES) glb_tiledefs[tile].wettile;
	if (wettile != TILE_INVALID)
	{
	    // Check to see if we crossed the threshold
	    if (getWetness(x, y) > glb_tiledefs[tile].wettransition)
	    {
		// New tile becomes un-wet.
		setWetness(x, y, 0);
		setTile(x, y, wettile);
		// A transition ocurred - trivially not steady!
		steadystate = false;

		// Check to see if someone fell into the water.
		if (wettile == TILE_WATER)
		{
		    MOB		*mob;
		    mob = getMob(x, y);
		    if (mob)
		    {
			msg_format("%S <fall> into the water!", mob);
		    }
		}
	    }
	}
	
	wettile = (TILE_NAMES) glb_tiledefs[tile].drytile;
	if (wettile != TILE_INVALID)
	{
	    // Check to see if we crossed the threshold
	    if (getWetness(x, y) < glb_tiledefs[tile].drytransition)
	    {
		// New tile becomes un-wet.
		setWetness(x, y, 0);
		setTile(x, y, wettile);

		// Check if someone got burned.
		if (wettile == TILE_FORESTFIRE || wettile == TILE_BRUSHFIRE)
		{
		    MOB		*mob;
		    mob = getMob(x, y);
		    if (mob)
		    {
			msg_format("%S <be> engulfed by the flames!", mob);
		    }
		}
		
		// A transition ocurred - trivially not steady!
		steadystate = false;
	    }
	}
    }

    // If transitions ocurred, our wetmap may have decresed in value
    // yet not yet be steady.
    if (!steadystate)
	return steadystate;

    FORALL_MAP(x, y)
    {
	if (getWetness(x, y) != oldwetmap(x, y))
	{
	    // See if this tile will ever transition
	    // If not, we don't care.
	    tile = getTile(x, y);
	    if (glb_tiledefs[tile].wettile != TILE_INVALID ||
		glb_tiledefs[tile].drytile != TILE_INVALID)
	    {
		// Not yet steady!
		return false;
	    }
	}
    }
    // No meaningful change in our wetmap1
    return true;
}

void
MAP::drownMobs(bool forcedrown)
{
    // allow even slow people to react.
    if (!forcedrown && spd_getphase() != PHASE_NORMAL)
	return;
    
    MOB		*mob, *next;
    TILE_NAMES	 tile;

    for (mob = myMobs; mob; mob = next)
    {
	next = mob->getNext();

	tile = getTile(mob->getX(), mob->getY());
	if (tile == TILE_WATER)
	{
	    msg_format("%S <drown>.", mob);
	    mob->applyDamage(0, 10);
	}
	else if (tile == TILE_BRUSHFIRE || tile == TILE_FORESTFIRE)
	{
	    msg_format("%S <burn>.", mob);
	    mob->applyDamage(0, 10);
	}
    }
}

int
MAP::calcScore() const
{
    int		x, y;
    TILE_NAMES	tile;

    int score = 0;

    // Very simple: number of floodable, but unflooded, squares.
    // We discount free squares from boulders.
    FORALL_MAP(x, y)
    {
	tile = getTile(x, y);
	// No points for water tiles.
	if (tile == TILE_WATER)
	    continue;

	// Score tiles that could turn to water, but didn't
	if (isWettable(x, y))
	{
	    score++;
	}

	// Score tiles that could burn, but didn't
	if (isBurnable(x, y))
	    score++;
    }

    // Halve score for each death
    if (!MOB::getAvatar())
	score /= 2;
    if (everHad(MOB_PRIMUS) && !findMob(MOB_PRIMUS))
	score /= 2;
    if (everHad(MOB_DEUCE) && !findMob(MOB_DEUCE))
	score /= 2;
    if (everHad(MOB_TREO) && !findMob(MOB_TREO))
	score /= 2;
	
    return score;
}

MAPDATA
MAP::buildConnectivity(int depth) const
{
    MAPDATA	connect(getWidth(), getHeight());
    int		x, y, dx, dy, nx, ny;
    
    connect.constant(-1);

    // Assign everyone to unique group
    FORALL_MAP(x, y)
    {
	connect.data(x, y) = y * 65536 + x;
    }
    
    FORALL_MAP(x, y)
    {
	if (!isPassable(x, y) || waterDepth(x,y) < depth)
	{
	    connect.data(x, y) = -1;
	    continue;
	}
	
	FORALL_8DIR(dx, dy)
	{
	    nx = x + dx;
	    ny = y + dy;
	    if (isPassable(nx, ny) && waterDepth(nx,ny) >= depth)
	    {
		connectClass(connect, x, y, nx, ny);
	    }
	}
    }

    return connect;
}

MAPDATA
MAP::buildWaterConnectivity() const
{
    MAPDATA	connect(getWidth(), getHeight());
    int		x, y, dx, dy, nx, ny;
    TILE_NAMES	tile;
    int		waterclass = -1;
    
    connect.constant(-1);

    // Assign everyone to unique group
    FORALL_MAP(x, y)
    {
	connect.data(x, y) = y * 65536 + x;
    }
    
    // Connect only "wettable" squares, meaning either is water,
    // has a transition, amd lacks item.
    FORALL_MAP(x, y)
    {
	if (!isWettable(x, y))
	{
	    connect.data(x, y) = -1;
	    continue;
	}

	// Only things with positive water sources are potentially
	// live.  Of those, only true-water gets full life.
	tile = getTile(x, y);
	if (glb_tiledefs[tile].wetpropagate <= 0)
	{
	    connect.data(x, y) = -1;
	    continue;
	}

	// If we are a water tile, add us to the water class
	if (getTile(x, y) == TILE_WATER)
	{
	    if (waterclass == -1)
	    {
		waterclass = findClass(connect, x, y);
	    }
	    connectToClass(connect, x, y, waterclass);
	}
	
	FORALL_4DIR(dx, dy)
	{
	    nx = x + dx;
	    ny = y + dy;
	    if (isWettable(nx, ny))
	    {
		tile = getTile(nx, ny);
		if (glb_tiledefs[tile].wetpropagate <= 0)
		{
		    continue;
		}

		connectClass(connect, x, y, nx, ny);
	    }
	}
    }

    // If waterclass is -1, nothing is valid as no water
    // Otherwise, do a sweep to convert everything into 0/1 for
    // if they are water connected
    if (waterclass == -1)
    {
	connect.constant(0);
	return connect;
    }

    MAPDATA result(getWidth(), getHeight());
    
    waterclass = findFromClass(connect, waterclass);
    FORALL_MAP(x, y)
    {
	if (findClass(connect, x, y) == waterclass)
	{
	    result.data(x, y) = 1;
	}
    }

    return result;
}

void
MAP::save(ostream &os) const
{
    os.write((const char *) &myWidth, sizeof(int));
    os.write((const char *) &myHeight, sizeof(int));

    myTiles.save(os);
    myTileFlags.save(os);
    myWetMap.save(os);
    
    myIntro.save(os);

    os.write((const char *) &myTotalRooms, sizeof(int));
    os.write((const char *) &myRooms, sizeof(int));

    MOB_NAMES mobname;
    FOREACH_MOB(mobname)
    {
	os.write((const char *) &myEverHad[mobname], sizeof(bool));
    }

    u8		c;
    MOB		*mob;
    ITEM	*item;

    for (mob = myMobs; mob; mob = mob->getNext())
    {
	c = 1;
	os.write((const char *) &c, 1);
	mob->save(os);
    }
    c = 0;
    os.write((const char *) &c, 1);
    
    for (item = myItems; item; item = item->getNext())
    {
	c = 1;
	os.write((const char *) &c, 1);
	item->save(os);
    }
    c = 0;
    os.write((const char *) &c, 1);
}

void
MAP::load(istream &is)
{
    u8			c;
    
    is.read((char *) &myWidth, sizeof(int));
    is.read((char *) &myHeight, sizeof(int));

    myTiles = MAPDATA(myWidth, myHeight);
    myTileFlags = MAPDATA(myWidth, myHeight);
    myWetMap = MAPDATA(myWidth, myHeight);

    myTiles.load(is);
    myTileFlags.load(is);
    myWetMap.load(is);

    myIntro.load(is);

    is.read((char *) &myTotalRooms, sizeof(int));
    is.read((char *) &myRooms, sizeof(int));

    MOB_NAMES mobname;
    FOREACH_MOB(mobname)
    {
	is.read((char *) &myEverHad[mobname], sizeof(bool));
    }

    // Load mobs...
    while (1)
    {
	is.read((char *) &c, 1);
	if (!c)
	    break;

	MOB		*mob;

	mob = MOB::load(is);

	addMob(mob);
    }

    // Load items
    while (1)
    {
	is.read((char *) &c, 1);
	if (!c)
	    break;

	ITEM		*item;

	item = ITEM::load(is);

	addItem(item);
    }
}

int
findLegend(const LEGENDLIST &legend, TILE_NAMES tilename, MOB_NAMES mobname, ITEM_NAMES itemname, int itemcount)
{
    int		i;

    for (i = 0; i < legend.entries(); i++)
    {
	if (legend(i).myTile != tilename)
	    continue;

	if (legend(i).myMob != mobname)
	    continue;
	
	if (legend(i).myItem != itemname)
	    continue;

	if (legend(i).myItemCount != itemcount)
	    continue;

	return i;
    }
    return -1;
}

static u8
addToLegend(LEGENDLIST &legend, u8 symbol, const char *name,
	    TILE_NAMES tile, MOB_NAMES mob, ITEM_NAMES item,
	    int itemcount)
{
    LEGEND	l(symbol, name, tile, mob, item, itemcount);
    legend.append(l);

    return symbol;
}

static u8
addToLegend(LEGENDLIST &legend, u8 symbol, TILE_NAMES tile, MOB_NAMES mob, ITEM_NAMES item, int itemcount)
{
    // Build the appropriate name...
    BUF		buf;
    int		inherit = -1;

    // Check for inheritance...
    if (mob != MOB_NONE || item != ITEM_NONE)
    {
	inherit = findLegend(legend, tile, MOB_NONE, ITEM_NONE, 0);
    }

    if (inherit >= 0)
    {
	buf.sprintf("(%c)", legend(inherit).getSymbol());
    }
    else
    {
	// Old fashion way.
	buf.sprintf("%s", glb_tiledefs[tile].legend);
    }

    if (mob != MOB_NONE)
    {
	buf.strcat(", ");
	buf.strcat(glb_mobdefs[mob].name);
    }

    if (item != ITEM_NONE)
    {
	int		j;

	for (j = 0; j < itemcount; j++)
	{
	    buf.strcat(", ");
	    buf.strcat(glb_itemdefs[item].name);
	}
    }

    return addToLegend(legend, symbol, buf.buffer(), tile, mob, item, itemcount);
}

static void
addToLegend(LEGENDLIST &legend, u8 symbol, TILE_NAMES tile)
{
    addToLegend(legend, symbol, tile, MOB_NONE, ITEM_NONE, 0);
}

static u8 glbFirstFree = 'A';

bool
legendHasSymbol(const LEGENDLIST &legend, u8 c)
{
    int		 i;
    for (i = 0; i < legend.entries(); i++)
    {
	if (legend(i).getSymbol() == c)
	    return true;
    }

    return false;
}

u8
findFreeSymbol(const LEGENDLIST &legend)
{
    while (legendHasSymbol(legend, glbFirstFree))
    {
	glbFirstFree++;
    }
    return glbFirstFree++;
}

static u8
addToLegend(LEGENDLIST &legend, TILE_NAMES tile, MOB_NAMES mob, ITEM_NAMES item, int itemcount)
{
    u8		symbol;

    symbol = findFreeSymbol(legend);
    return addToLegend(legend, symbol, tile, mob, item, itemcount);
}

u8
MAP::mapToLegend(int x, int y, LEGENDLIST &legend) const
{
    int		i;
    MOB		*mob;
    ITEM	*item;
    MOB_NAMES	mobname;
    ITEM_NAMES	itemname;
    TILE_NAMES	tilename;
    int		itemcount = 0;

    tilename = getTile(x, y);

    mobname = MOB_NONE;
    mob = getMob(x, y);
    if (mob)
	mobname = mob->getDefinition();

    // Some day we should add support for item stacks.
    itemname = ITEM_NONE;
    item = getItem(x, y);
    if (item)
    {
	itemname = item->getDefinition();
	// Because it is forbidden to place boulders on top of rocks,
	// we can just merge different rock types into one rock type
	// since it doesn't matter in game which one it is.
	ITEMLIST		itemlist;

	getItemList(itemlist, x, y);
	
	itemcount = 0;
	for (i = 0; i < itemlist.entries(); i++)
	{
	    itemcount += itemlist(i)->getStackCount();
	}
    }

    // First see if we are already a legend.
    i = findLegend(legend, tilename, mobname, itemname, itemcount);

    if (i >= 0)
    {
	return legend(i).getSymbol();
    }

    // Failure, create a new legend to return
    return addToLegend(legend, tilename, mobname, itemname, itemcount);
}

BUF
MAP::saveToText() const
{
    LEGENDLIST		legend;
    int			x, y, i;
    BUF			result;

    // Prebuild our usual symbols.
    addToLegend(legend, '^', TILE_ICEMOUNTAIN);
    addToLegend(legend, 'V', TILE_MOUNTAIN);
    addToLegend(legend, '=', TILE_BRIDGE);
    addToLegend(legend, '#', TILE_WALL);
    addToLegend(legend, '.', TILE_PATH);
    addToLegend(legend, ',', TILE_MUD);
    addToLegend(legend, ':', TILE_FLOOR);
    addToLegend(legend, ';', TILE_SNOWYPATH);
    addToLegend(legend, ' ', TILE_GRASS);
    addToLegend(legend, '&', TILE_FOREST);
    addToLegend(legend, '"', TILE_FIELD);
    addToLegend(legend, '!', TILE_FORESTFIRE);
    addToLegend(legend, '%', TILE_BRUSHFIRE);
    addToLegend(legend, '~', TILE_WATER);
    addToLegend(legend, '0', TILE_GRASS, MOB_NONE, ITEM_BOULDER, 1);
    addToLegend(legend, 'O', TILE_FOREST, MOB_NONE, ITEM_BOULDER, 1);
    addToLegend(legend, '*', TILE_GRASS, MOB_NONE, ITEM_LARGEROCK, 1);
    addToLegend(legend, '}', TILE_FOREST, MOB_NONE, ITEM_LARGEROCK, 1);
    addToLegend(legend, '+', TILE_GRASS, MOB_NONE, ITEM_SMALLROCK, 1);
    addToLegend(legend, '{', TILE_FOREST, MOB_NONE, ITEM_SMALLROCK, 1);

    // Reset our free list.
    glbFirstFree = 'A';

    result.strcat("# Fatherhood Map\n");
    result.strcat("\n");

    // Map beginning
    result.append('+');
    for (x = 0; x < getWidth(); x++)
	result.append('-');
    result.append('+');
    result.append('\n');

    for (y = 0; y < getHeight(); y++)
    {
	result.append('|');

	for (x = 0; x < getWidth(); x++)
	{
	    result.append(mapToLegend(x, y, legend));
	}
	
	result.append('|');
	result.append('\n');
    }

    // Map Ending
    result.append('+');
    for (x = 0; x < getWidth(); x++)
	result.append('-');
    result.append('+');
    result.append('\n');

    // Now output the legend.
    result.strcat("\n# Legend:\n");
   
    for (i = 0; i < legend.entries(); i++)
    {
	BUF		buf;

	buf.sprintf("%c: %s\n", legend(i).getSymbol(), legend(i).getDescr());
	result.strcat(buf);
    }

    return result;
}
