/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        map.h ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#ifndef __map__
#define __map__

#include "glbdef.h"

#include "pos.h"
#include "ptrlist.h"
#include "mapdata.h"
#include "mob.h"
#include "item.h"

#include <iostream>
using namespace std;

class LEGEND;
typedef PTRLIST<LEGEND> LEGENDLIST;

#define MAPFLAG_FOV 1
#define MAPFLAG_MAP 2
#define MAPFLAG_ITEM 4
#define MAPFLAG_MOB 8

class MAP
{
public:
		 MAP();
		~MAP();

    // Computes distance between two points according to this map's
    // distance metric.
    static int	 dist(int x1, int y1, int x2, int y2);		
    static int	 dist(const MOB *a, const MOB *b);
    static int	 dist(const MOB *a, int x, int y);

    // Returns true if the points are orthogonal or diagonal.
    static bool	 isLinedUp(int x1, int y1, int x2, int y2);
    static bool	 isLinedUp(const MOB *a, const MOB *b);
    static bool	 isLinedUp(const MOB *a, int x, int y);

    // Unlike the copy, this will copy all maps reachable from this
    // map.
    MAP		*copyAll() const;		
    MAP		*copy() const;

    // Searches for the avatar in all linked maps, setting the MOB::avatar
    // accordingly.  Sets to 0 if no avatar present.
    void	locateAvatar(MOB_NAMES avatarname);
    void	locateAvatarLocal(MOB_NAMES avatarname);

    bool	everHad(MOB_NAMES mob) const
		{ return myEverHad[mob]; }

    void	doMoveNPC();		

    bool 	flowLiquids(bool forceflow=false);
    void	drownMobs(bool forcedrown=false);
    
    int		getWidth() const { return myWidth; }
    int		getHeight() const { return myHeight; }
    bool	isValidCoord(int x, int y) const;
    bool	isValidCoord(POS p) const
		{ return isValidCoord(p.x, p.y); }

    TILE_NAMES	getTile(int x, int y) const;
    TILE_NAMES	getTile(POS p) const
		{ return getTile(p.x, p.y); }
    void	setTile(int x, int y, TILE_NAMES tile);
    void	setTile(POS p, TILE_NAMES tile)
		{ setTile(p.x, p.y, tile); }

    // It is valid to address this out of bounds, such
    // calls are ignored.
    // Reads and controls the wetness of every tile.
    // Note: Does not transition wetness.
    int		getWetness(int x, int y) const;
    void	setWetness(int x, int y, int value);
    void	addWetness(int x, int y, int amount);

    int		waterDepth(int x, int y) const;
    int		waterDepth(POS p) const { return waterDepth(p.x, p.y); }

    int		wallDist(POS p) const;

    // True if tile is water or can transition
    bool	isWettable(int x, int y) const;
    bool	isBurnable(int x, int y) const;

    // Determines how well you did.    
    int		calcScore() const;

    bool	findNextSymbol(char symbol, int &x, int &y) const;
    bool	findTile(TILE_NAMES tile, int &x, int &y, bool avoidmob = false) const;

    bool	findMoveable(int &x, int &y, bool avoidmob = false) const;
    bool	findMoveableUnexplored(int &x, int &y) const;
    bool	findCloseMoveableUnexplored(int &x, int &y,
				    int cx, int cy) const;

    bool	isPassable(int x, int y) const;
    bool	isPassable(POS p) const
		{ return isPassable(p.x, p.y); }
    bool	isTransparent(int x, int y) const;

    bool	isMapped(int x, int y) const;
    bool	isFOV(int x, int y) const;
    bool	isFOV(POS p) const { return isFOV(p.x, p.y); }

    // Determine if you can walk from p1 to p2, 8 ways, only passable
    // terrain
    bool	 isConnected(POS p1, POS p2, int depth = 0) const;

    ITEM	*getItem(int x, int y) const;
    bool	 hasItem(int x, int y) const;
    bool	 hasItem(POS p) const
		 { return hasItem(p.x, p.y); }
    void	 getItemList(ITEMLIST &list, int x, int y) const;

    MOB		*getMob(int x, int y) const;
    MOB		*getMob(POS p) const
		 { return getMob(p.x, p.y); }
    bool	 hasMob(int x, int y) const;
    bool	 hasMob(POS p) const
		 { return hasMob(p.x, p.y); }
    int		 getNumMobs() const;
    int		 getAllMobs(PTRLIST<MOB *> &moblist, int x, int y) const;
    MOB		*findMob(MOB_NAMES def) const;

    MOB		*traceBullet(int x, int y, int r, int dx, int dy) const;
    void	 displayBullet(int x, int y, int r, int dx, int dy, u8 sym, ATTR_NAMES attr) const;

    void	 fireball(MOB *src, int x, int y, int rad,
			DPDF damage, u8 symbol, ATTR_NAMES attr);

    void	 describeSquare(int x, int y, bool isblind = false) const;

    BUF		 getTerrainName(int x, int y) const;

    void	 addItem(ITEM *item);
    void	 removeItem(ITEM *item);

    void	 addMob(MOB *mob);
    void	 removeMob(MOB *mob);

    int	 	 getDepth()  const { return (myTotalRooms / 40) + 1; }
    int		 getRooms() const { return myRooms; }

    // Finds the two points farthest apart by walking.
    // This is increasingly misnamed as it is generalized for n points.
    void	 getDiameter(POSLIST &list, int npt, bool quiet = false) const;

    // Finds a position closest to p yet at least dist away
    // from the walls.
    POS		 findSetBack(POS p, int dist) const;

    // Returns point farthest away from a.
    POS		 findDeepest(POS a) const;

    // Returns deepest from a list of coordinates.
    // If list empty, random cooridinate is used as a source
    // for findDeepest.
    POS		 findDeepest(const POSLIST &list) const;

    // Returns distance walking between p1 and p2.
    // Caches a map wrt to p1, so make that constant if possible.
    int		 getDist(POS p1, POS p2) const;

    MAPDATA	 getDistMap(POS p) const;

    // Marks everything walkable from p or 8 way adjacent to such
    // as mapped.
    void	 mapConnected(POS p);

    // Resets any cached maps that might depend on topology
    void	 clearAllCaches();

    MAPDATA	 buildConnectivity(int depth) const;
    MAPDATA	 buildWaterConnectivity() const;
    MAPDATA	 buildWaterDepth() const;
    MAPDATA	 buildWallDist() const;
    MAPDATA	 buildDistMap(POS p) const;

    // Redraws to the given window.
    // cx & cy are the center of the map, x-w and y-h are the
    // window to draw into in screen coords.
    void	redraw(int x, int y, int w, int h,
			int cx, int cy) const;

    // Redraws with the last known window.
    void	re_redraw() const;

    // Fades the map to black.
    void	fadeToBlack() const;

    // Constructs a proper map for the given depth.
    void	build(MAPSIZE_NAMES size, int numriver, int numfire);

    MAP		*getDownMap() const;
    MAP		*getUpMap() const;

    // Used for loading...
    void	 setDownMap(MAP *down);
    void	 setUpMap(MAP *down);
    MAP		*peekDownMap() const { return myDownMap; }
    MAP		*peekUpMap() const { return myDownMap; }

    BUF		 getIntroText() const { return myIntro; }
    BUF		 getInitialText() const { return myOrigState; }

    MOB		*getMobHead() const { return myMobs; }
    ITEM	*getItemHead() const { return myItems; }

    BUF		 getMapName() const { return myMapName; }
    void	 setMapName(const char *name);

    // Applies given symbol to given x,y coord according to given legend.
    void	parseLegend(int x, int y, const LEGENDLIST &list,
			    u8 symbol);

    // Loads a map from the given .map file
    void	loadMap(const char *fname);

    void	clear(TILE_NAMES tile);
    void	clearFlag(int flag, bool val);

    // Sets the FOV flag for everything in sight of cx & cy.
    void	buildFOV(int cx, int cy);

    // Checks all 9 squares around sx-sy for LOS.  This lets you
    // look around corners.
    bool	hasLOSLoose(int sx, int sy, int ex, int ey) const;

    // Return strue if (sx, sy) has line of sight to (ex, ey)
    bool	hasLOS(int sx, int sy, int ex, int ey) const;

    void	load(istream &is);
    void	save(ostream &os) const;
    
    bool	 getFlag(int x, int y, int flag) const;
    void	 setFlag(int x, int y, int flag, bool val) const;
    void	 setFlag(POS p, int flag, bool val) const
		 { setFlag(p.x, p.y, flag, val); }

    // Transforms this world into a text world.
    BUF		 saveToText() const;
    u8		 mapToLegend(int x, int y, LEGENDLIST &legend) const;

protected:
    int		 myWidth, myHeight;

    // These cache the last displayed values of the map so we can
    // redraw it at our leisure.
    mutable int	 myLastX, myLastY, myLastW, myLastH, myLastMX, myLastMY;

    // Raw array of tiles & flags
    MAPDATA	 myTiles;

    // Must be mutable because we have caches stored here.
    mutable MAPDATA	 myTileFlags;

    // How wet each square is.
    MAPDATA	 myWetMap;

    mutable MAPDATA	 myConnectivity;
    mutable int	 	 myConnectivityDepth;

    mutable MAPDATA	 myWaterDepth;
    mutable bool	 myWaterDepthValid;

    mutable MAPDATA	 myWallDist;
    mutable bool	 myWallDistValid;

    // Cached distance map along with which point it is a distance
    // with respect to.
    mutable PTRLIST<MAPDATA> 	myDistMaps;
    mutable POSLIST		myDistMapPos;

    mutable POSLIST		myMoveablePos;

    // List of all of our mobs.
    MOB		*myMobs;
    MOB		*myMobSafeNext;

    // List of all our items.
    ITEM	*myItems;

    int		 myTotalRooms;
    int		 myRooms;

    bool	 myEverHad[NUM_MOBS];

    BUF		 myMapName;
    BUF		 myIntro;

    BUF		 myOrigState;

    mutable MAP	*myDownMap, *myUpMap;
};

#endif

